#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   converter.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   TensorRT converter module
"""

# pylint: disable=E1102,R0902,R0913,R0914,R0915

import os
import shutil
import subprocess
import time
from typing import Dict, List, Optional, Tuple

from datature.nexus import config, error, models
from datature.nexus.api.project import Project
from datature.nexus.api.types import OperationStatusOverview

from .logger import trt_logger
from .params import ConversionParams, DockerParams

try:
    import pynvml
except ModuleNotFoundError as import_exc:
    raise ModuleNotFoundError(
        "nvidia-ml-py not installed! "
        "Please visit https://developers.datature.io/docs/tensorrt-export "
        "to learn how to set up your environment."
    ) from import_exc

try:
    import docker
except ModuleNotFoundError as import_exc:
    raise ModuleNotFoundError(
        "Docker not installed! "
        "Please visit https://developers.datature.io/docs/tensorrt-export "
        "to learn how to set up your environment."
    ) from import_exc


class Converter:
    """The Converter class provides an interface to the TensorRT Converter."""

    _min_driver_version: str = "530.00"
    _min_cuda_version: str = "11.8"
    _model_directory: str = ".datature_models"

    def __init__(
        self,
        docker_client,
        api_client,
        docker_params: Optional[DockerParams] = DockerParams(),
        conversion_params: Optional[ConversionParams] = ConversionParams(),
    ) -> None:
        """
        Instantiates the TensorRT Converter.

        Args:
            docker_client (docker.client.DockerClient): The Docker client instance.
            api_client (docker.api.client.APIClient): The Docker API client instance.
            conversion_docker_image (str):
                The Docker image name used for TensorRT conversion,
                defaults to "nvcr.io/nvidia/tensorflow:23.04-tf2-py3".
                The list of Docker images can be found here:
                https://catalog.ngc.nvidia.com/orgs/nvidia/containers/tensorflow/tags
            device (int): GPU id to use for conversion, defaults to 0.
            timeout (int):
                The maximum time in seconds to wait for the conversion process to complete,
                defaults to 1800.
        """

        self._docker_client = docker_client
        self._api_client = api_client
        self._docker_params = docker_params
        self._conversion_params = conversion_params

        try:
            self._device_request = docker.types.DeviceRequest(
                device_ids=[f"{self._docker_params.device}"], capabilities=[["gpu"]]
            )
        except docker.errors.DockerException as device_exc:
            raise ConnectionError(
                f"Could not detect device {self._docker_params.device}. "
                "Please ensure that your system contains an NVIDIA GPU "
                "or NVIDIA-compatible edge device, such as a Jetson."
            ) from device_exc
        self._exec_id = ""
        self._logs = None
        self._cont = None

    def setup(self) -> None:
        """Checks the system environment and sets up the conversion Docker image."""
        os_name = os.uname().sysname

        if os_name == "Darwin":
            raise SystemError(
                "TensorRT conversion is not supported on MacOS. "
                "Please refer to our documentation for supported platforms: "
                "https://developers.datature.io/docs/tensorrt-export"
            )

        if os_name == "Windows":
            raise SystemError(
                f"TensorRT conversion is not supported natively on {os_name}. "
                "You will need to install Windows Subsystem for Linux (WSL). "
                "Please refer to our documentation for supported platforms and "
                "installation guides: https://developers.datature.io/docs/tensorrt-export"
            )

        if os_name == "Linux" and os.path.exists("/proc/sys/kernel/osrelease"):
            with open(
                "/proc/sys/kernel/osrelease", "r", encoding="utf-8"
            ) as os_release:
                if "WSL2" in os_release.read():
                    trt_logger.info("OS: Windows Subsystem for Linux 2 (WSL2)")
                else:
                    trt_logger.info("OS: %s", os_name)

        else:
            raise SystemError(
                f"Your OS {os_name} is unsupported. Please refer to "
                "https://developers.datature.io/docs/tensorrt-export "
                "for supported platforms."
            )

        gpu_info = self._get_gpu_info()

        if gpu_info["driver_version"] is None:
            raise SystemError("Could not determine GPU driver version")

        if str(gpu_info["driver_version"]) < self._min_driver_version:
            raise SystemError(
                f"GPU driver version is less than {self._min_driver_version}, "
                "please upgrade to a newer version. Please refer to "
                "https://developers.datature.io/docs/tensorrt-export "
                "for more information."
            )

        if gpu_info["cuda_version"] is None:
            raise SystemError("Could not determine CUDA version")

        if str(gpu_info["cuda_version"]) < self._min_cuda_version:
            raise SystemError(
                f"CUDA version is less than {self._min_cuda_version}, "
                "please upgrade to a newer version. Please refer to "
                "https://developers.datature.io/docs/tensorrt-export "
                "for more information."
            )

        # environment check: docker and client machine compatibility
        candidate = self._docker_client.images.list(
            filters={"reference": self._docker_params.conversion_docker_image}
        )

        if len(candidate) == 0:
            trt_logger.info(
                "Pulling Docker image %s", self._docker_params.conversion_docker_image
            )
            self._docker_client.images.pull(self._docker_params.conversion_docker_image)

    def convert(
        self,
        model_path: str,
        output_path: str = "",
    ) -> None:
        """
        Converts an ONNX model to TensorRT format.

        Args:
            model_path (str): The path to the saved model directory.
            output_path (str): The path to the output directory.
        """
        try:
            commands = self._get_commands(model_path, output_path)
            self._start_docker_container()
            self._exec_id = self._api_client.exec_create(
                container=self._cont.id,
                cmd=commands,
                environment={
                    "POLYGRAPHY_AUTOINSTALL_DEPS": (
                        1 if self._conversion_params.autoinstall_deps else 0
                    ),
                    "POLYGRAPHY_INTERNAL_CORRECTNESS_CHECKS": (
                        1 if self._conversion_params.internal_correctness_checks else 0
                    ),
                },
            )["Id"]
            self._logs = self._api_client.exec_start(
                self._exec_id, stream=True, detach=False
            )

            self._print_logs()

            if self.get_status() != 0:
                raise error.Error(
                    "TensorRT conversion was not successful, please read "
                    "the error logs printed above or visit "
                    "https://developers.datature.io/docs/tensorrt-export for more information."
                )

            trt_logger.info(
                "TensorRT model conversion successful, model is saved at %s", model_path
            )
            trt_logger.info(
                "Run `with trt.inference_server(YOUR_MODEL_FOLDER):` to spin up "
                "NVIDIA Triton Inference Server for local inference. "
                "Refer to https://developers.datature.io/docs/tensorrt-export "
                "on how to make predictions with your own data."
            )
        except KeyboardInterrupt:
            trt_logger.warning("Keyboard interrupt detected")
        except Exception as exc:
            raise error.Error(exc) from exc
        finally:
            self._cleanup()

    def create_export(
        self,
        project: Project,
        artifact_id: str,
    ) -> str:
        """
        Exports an ONNX model from Datature Nexus given a Project instance
        and an artifact ID, and returns the model ID.

        Args:
            project (Project): The Project instance.
            artifact_id (str): The artifact ID of the model to export.

        Returns:
            str: The exported model ID.

        Raises:
            InternalServerError: If the artifact export encounters an error.
            BadRequestError: If the artifact export has previously been cancelled.
        """
        assert isinstance(project, Project)
        assert isinstance(artifact_id, str)

        target_format = "TensorRT"
        export_options = {"format": target_format}
        is_exported = False

        try:
            model = project.requester.POST(
                f"/projects/{project.project_id}/artifacts/{artifact_id}/hiddenExports",
                request_body=export_options,
                response_type=models.ArtifactModel,
            )
            model_id = model.id
        except error.BadRequestError as exc:
            raise error.BadRequestError(
                "Your model architecture is not supported for TensorRT conversion. "
                "Please refer to our documentation for supported model architectures: "
                "https://developers.datature.io/docs/tensorrt-export"
            ) from exc
        except error.InternalServerError as exc:
            if exc.detail["code"] == "ConflictError":
                trt_logger.info("Model has already been exported, retrieving export")
                is_exported = True
                model_id = [
                    model
                    for model in project.artifacts.list_exported_models(artifact_id)
                    if model.format == target_format
                ][-1].id
            else:
                raise error.InternalServerError(
                    "Artifact export encountered an error, "
                    "please contact our engineers at support@datature.io for support."
                ) from exc

        if not is_exported:
            target_export = [
                export
                for export in project.artifacts.list_exported_models(artifact_id)
                if export.id == model_id
            ][-1]

            while target_export.status != OperationStatusOverview.FINISHED.value:
                if target_export.status == OperationStatusOverview.ERRORED.value:
                    raise error.InternalServerError(
                        "Artifact export encountered an error, "
                        "please contact our engineers at support@datature.io for support."
                    )

                if target_export.status == OperationStatusOverview.CANCELLED.value:
                    raise error.BadRequestError(
                        "Artifact export has previously been cancelled by user, "
                        "please start a new export."
                    )

                time.sleep(config.OPERATION_LOOPING_DELAY_SECONDS)
                target_export = [
                    export
                    for export in project.artifacts.list_exported_models(artifact_id)
                    if export.id == model_id
                ][-1]
        return model_id

    def download_export(self, project: Project, model_id: str) -> str:
        """
        Downloads the exported model from Nexus to a local directory given a model ID,
        and returns the path to the downloaded model.

        Args:
            project (Project): The Project instance.
            model_id (str): The model ID of the model to download.

        Returns:
            str: The path to the downloaded model.

        Raises:
            FileNotFoundError: If the path to the downloaded ONNX model cannot be found.
        """
        download_details = project.artifacts.download_exported_model(
            model_id, self._model_directory
        )
        model_path = os.path.join(
            download_details.download_path, download_details.model_filename
        )

        if not os.path.exists(model_path):
            raise FileNotFoundError(
                "There was a problem locating your model file, "
                "please ensure that you have provided a valid directory "
                "and re-run the .convert() function again"
            )

        return model_path

    def _start_docker_container(self) -> None:
        """Starts docker container used for TensorRT conversion."""
        working_dir = os.getcwd()

        self._docker_client.images.get(self._docker_params.conversion_docker_image)
        container = self._docker_client.containers.run(
            self._docker_params.conversion_docker_image,
            stdin_open=True,
            auto_remove=True,
            detach=True,
            volumes={
                working_dir: {
                    "bind": self._docker_params.workspace_path,
                    "mode": self._docker_params.dir_perms,
                }
            },
            device_requests=[self._device_request],
        )
        self._cont = self._docker_client.containers.get(container.id)
        self._cont.start()

        start = time.time()
        while self._cont.status != "running":
            self._cont = self._docker_client.containers.get(container.id)

            if self._cont.status == "running":
                pass
            elif self._cont.status == "paused":
                self._cont.unpause()
            elif self._cont.status == "exited":
                self._cont.start()
            elif self._cont.status == "restarting":
                pass
            else:
                raise error.Error(f"Error starting container: {container.id}")

            time.sleep(1)

            if int(time.time() - start) > self._conversion_params.timeout:
                raise TimeoutError(
                    "Docker container startup timed out. "
                    "Please refer to our documentation "
                    "https://developers.datature.io/docs/tensorrt-export for more details, "
                    "or contact our engineers at support@datature.io for assistance."
                )
        trt_logger.info("Docker container running")

    def _get_commands(
        self,
        model_path: str,
        output_path: str = "",
    ) -> List[str]:
        """
        Constructs a command string for running a model conversion script.

        Args:
            precision (str): The quantization precision, either "FP32" or "INT8".
            model_path (str): The path to the saved model directory.
            output_path (str): The path to the output directory.

        Returns:
            List[str]: A list of command strings for running the model conversion script.

        Raises:
            ValueError: If the precision is "INT8", as it is currently not yet supported.
            ValueError: If the mode is not "FP32" or "FP16".
            ValueError: If the model path is not a valid ONNX model.
        """
        if self._conversion_params.precision == "INT8":
            raise ValueError(
                "INT8 quantization is currently not yet supported."
                "Please contact our engineers at support@datature.io "
                "if you require a custom conversion."
            )

        if self._conversion_params.precision not in ["FP32", "FP16"]:
            raise ValueError(
                "Invalid mode. Must be either 'FP32' or 'FP16'."
                "INT8 quantization is currently not yet supported."
                "Please contact our engineers at support@datature.io "
                "if you require a custom conversion."
            )

        if not os.path.exists(model_path) or not model_path.endswith(".onnx"):
            raise ValueError(
                "Invalid model path."
                "Please ensure that you provide the correct path to your ONNX model."
                "Refer to our documentation for supported model files: "
                "https://developers.datature.io/docs/tensorrt-export"
            )

        os.makedirs(self._conversion_params.log_dir, exist_ok=True)
        log_file = os.path.join(
            self._conversion_params.log_dir, f"{int(time.time())}.log"
        )
        trt_logger.info("Verbose logs will be saved at: %s", log_file)

        commands = ["sh", "-c", "pip3 install -qqq colored"]

        check_commands = ""
        sanitize_commands = ""
        conversion_experimental_commands = ""

        if self._conversion_params.experimental:
            (
                check_commands,
                sanitize_commands,
                conversion_experimental_commands,
            ) = self._get_experimental_commands(
                model_path,
                check_commands,
                sanitize_commands,
                conversion_experimental_commands,
            )

        precision_flag = "--fp16" if self._conversion_params.precision == "FP16" else ""
        convert_verbose_flag = (
            "--verbosity VERBOSE" if self._conversion_params.verbose else ""
        )
        conversion_commands = (
            f" && polygraphy convert {model_path}"
            f" --model-type onnx"
            f" -o {output_path}"
            f" --convert-to trt"
            f" {precision_flag}"
            f" --builder-optimization-level {self._conversion_params.builder_optimization_level}"
            f" --precision-constraints {self._conversion_params.precision_constraints}"
            f" {convert_verbose_flag}"
            f" --log-file {log_file}"
        )
        commands[-1] += check_commands
        commands[-1] += sanitize_commands
        commands[-1] += conversion_commands
        commands[-1] += conversion_experimental_commands
        return commands

    def _get_experimental_commands(
        self,
        model_path: str,
        check_commands: str,
        sanitize_commands: str,
        conversion_experimental_commands: str,
    ) -> Tuple[str, str, str]:
        """
        Constructs a command string for running experimental conversion settings.

        Args:
            check_commands (str): The command string for running `polygraphy check lint`.
            sanitize_commands (str): The command string for running `polygraphy surgeon sanitize`.
            conversion_experimental_commands (str):
                The command string for adding experimental conversion settings.

        Returns:
            Tuple[str, str, str]: A tuple of command strings for running the
                experimental conversion settings.
        """
        experimental_params = self._conversion_params.experimental
        check_params = experimental_params.check_params
        sanitize_params = experimental_params.sanitize_params
        if check_params.enabled:
            trt_logger.warning(
                "[EXPERIMENTAL] `polygraphy check lint` has been enabled."
                " This is an experimental setting and conversion may fail."
                " This check will be run before TensorRT conversion to"
                " topologically lint ONNX model to find faulty nodes in the graph."
            )
            output_json_path_flag = (
                f"--output_json_path {check_params.output_json_path}"
            )
            provider = f"--provider {check_params.provider}"
            check_verbosity_flag = "--verbosity VERBOSE" if check_params.verbose else ""
            check_commands = (
                f" && polygraphy check lint {model_path}"
                f" {output_json_path_flag}"
                f" {provider}"
                f" {check_verbosity_flag}"
            )

        if sanitize_params.enabled:
            trt_logger.warning(
                "[EXPERIMENTAL] `polygraphy surgeon sanitize` has been enabled."
                " This is an experimental setting and conversion may fail."
                " ONNX graph surgeon will be run before TensorRT conversion to"
                " clean up and optimize input shapes in an ONNX model."
            )
            root_dir, file_name = os.path.split(model_path)
            output_path = (
                sanitize_params.output_model_path
                if sanitize_params.output_model_path
                else os.path.join(root_dir, f"sanitized_{file_name}")
            )
            cleanup_flag = "--cleanup" if sanitize_params.cleanup else ""
            toposort_flag = "--toposort" if sanitize_params.toposort else ""
            no_shape_inference_flag = (
                "--no-shape-inference" if sanitize_params.no_shape_inference else ""
            )
            force_fallback_shape_inference_flag = (
                "--force-fallback-shape-inference"
                if sanitize_params.force_fallback_shape_inference
                else ""
            )
            fold_constants_flag = (
                "--fold-constants" if sanitize_params.fold_constants else ""
            )
            partitioning_flag = (
                f"--partitioning {sanitize_params.partitioning}"
                if sanitize_params.partitioning
                else ""
            )
            no_fold_shapes_flag = (
                "--no-fold-shapes" if sanitize_params.no_fold_shapes else ""
            )
            no_per_pass_shape_inference_flag = (
                "--no-per-pass-shape-inference"
                if sanitize_params.no_per_pass_shape_inference
                else ""
            )
            sanitize_commands = (
                f" && polygraphy surgeon sanitize {model_path}"
                f" --output {output_path}"
                f" {cleanup_flag}"
                f" {toposort_flag}"
                f" {no_shape_inference_flag}"
                f" {force_fallback_shape_inference_flag}"
                f" {fold_constants_flag}"
                f" {partitioning_flag}"
                f" {no_fold_shapes_flag}"
                f" {no_per_pass_shape_inference_flag}"
            )
            model_path = output_path

        docs_reference = (
            "Please refer to"
            " https://docs.nvidia.com/deeplearning/tensorrt/polygraphy/docs/backend/trt/config.html"
            " for more information."
        )
        sparse_weights_flag = ""
        version_compatible_flag = ""
        error_on_timing_cache_miss_flag = ""
        load_timing_cache_flag = ""
        save_timing_cache_flag = ""
        disable_compilation_cache_flag = ""
        load_tactics_flag = ""
        save_tactics_flag = ""

        if experimental_params.sparse_weights:
            sparse_weights_flag = "--sparse-weights"
            trt_logger.warning(
                "[EXPERIMENTAL] `%s` has been enabled."
                " This is an experimental setting and conversion may fail."
                " This will enable optimizations for sparse weights in TensorRT. %s",
                sparse_weights_flag,
                docs_reference,
            )

        if experimental_params.version_compatible:
            version_compatible_flag = "--version-compatible"
            trt_logger.warning(
                "[EXPERIMENTAL] `%s` has been enabled."
                " This is an experimental setting and conversion may fail."
                " This will build an engine designed to be"
                " forward TensorRT version compatible. %s",
                version_compatible_flag,
                docs_reference,
            )

        if experimental_params.error_on_timing_cache_miss:
            error_on_timing_cache_miss_flag = "--error-on-timing-cache-miss"
            trt_logger.warning(
                "[EXPERIMENTAL] `%s` has been enabled."
                " This is an experimental setting and conversion may fail."
                " The conversion process will emit errors when a tactic being timed"
                " is not present in the timing cache. %s",
                error_on_timing_cache_miss_flag,
                docs_reference,
            )

        if experimental_params.load_timing_cache:
            load_timing_cache_flag = (
                f"--load-timing-cache {experimental_params.load_timing_cache}"
            )
            trt_logger.warning(
                "[EXPERIMENTAL] `%s` has been enabled."
                " This is an experimental setting and conversion may fail."
                " This will load the specified tactic timing cache [%s]"
                " used to speed up the TensorRT engine building process. %s",
                load_timing_cache_flag,
                experimental_params.load_timing_cache,
                docs_reference,
            )

        if experimental_params.save_timing_cache:
            save_timing_cache_flag = (
                f"--save-timing-cache {experimental_params.save_timing_cache}"
            )
            trt_logger.warning(
                "[EXPERIMENTAL] `%s` has been enabled."
                " This is an experimental setting and conversion may fail."
                " This will enable saving of tactic timing cache to file [%s]. %s",
                save_timing_cache_flag,
                experimental_params.save_timing_cache,
                docs_reference,
            )

        if experimental_params.disable_compilation_cache:
            disable_compilation_cache_flag = "--disable-compilation-cache"
            trt_logger.warning(
                "[EXPERIMENTAL] `%s` has been enabled."
                " This is an experimental setting and conversion may fail."
                " This will disable caching of JIT-compiled code. %s",
                disable_compilation_cache_flag,
                docs_reference,
            )

        if experimental_params.load_tactics:
            load_tactics_flag = f"--load-tactics {experimental_params.load_tactics}"
            trt_logger.warning(
                "[EXPERIMENTAL] `%s` has been enabled."
                " This is an experimental setting and conversion may fail."
                " This will load specified tactic replay file [%s], which will"
                " override tactics in TensorRT's default selections. %s",
                load_tactics_flag,
                experimental_params.load_tactics,
                docs_reference,
            )

        if experimental_params.save_tactics:
            save_tactics_flag = f"--save-tactics {experimental_params.save_tactics}"
            trt_logger.warning(
                "[EXPERIMENTAL] `%s` has been enabled."
                " This is an experimental setting and conversion may fail."
                " This will save tactics selected by TensorRT to a"
                " specified JSON file [%s]. %s",
                save_tactics_flag,
                experimental_params.save_tactics,
                docs_reference,
            )

        conversion_experimental_commands = (
            f" {sparse_weights_flag}"
            f" {version_compatible_flag}"
            f" {error_on_timing_cache_miss_flag}"
            f" {load_timing_cache_flag}"
            f" {save_timing_cache_flag}"
            f" {disable_compilation_cache_flag}"
            f" {load_tactics_flag}"
            f" {save_tactics_flag}"
        )
        return check_commands, sanitize_commands, conversion_experimental_commands

    def _get_gpu_info(self) -> Dict[str, str]:
        """
        Returns the GPU information of the system.

        Returns:
            dict: A dictionary containing the GPU name, driver version, and CUDA version.
        """
        try:
            pynvml.nvmlInit()
            driver_version = pynvml.nvmlSystemGetDriverVersion()
            if isinstance(driver_version, bytes):
                driver_version = driver_version.decode("utf-8")
            cuda_version = pynvml.nvmlSystemGetCudaDriverVersion()

            device_count = pynvml.nvmlDeviceGetCount()
            if device_count == 0:
                raise error.Error("No NVIDIA GPU found")
            handle = pynvml.nvmlDeviceGetHandleByIndex(0)
            gpu_name = str(pynvml.nvmlDeviceGetName(handle))
            pynvml.nvmlShutdown()

            return {
                "gpu_name": gpu_name,
                "driver_version": driver_version,
                "cuda_version": cuda_version,
            }
        except subprocess.CalledProcessError as exc:
            raise error.Error(f"Error running nvidia-settings: {exc.stderr}") from exc
        except pynvml.NVMLError as exc:
            raise error.NotFoundError(
                f"{exc}\nError: TensorRT conversion requires pre-installed NVIDIA drivers.\n"
                "If the necessary NVIDIA drivers are already installed, you might lack sufficient permissions to access them.\n"
            ) from exc
        except error.Error as exc:
            raise exc from exc

    def _cleanup(self):
        """Stops Docker container and removes temporary ONNX exported models."""
        trt_logger.info("Cleaning up TensorRT conversion resources...")
        if self._cont.status != "exited":
            self._cont.stop()
            self._docker_client.containers.prune()

        if os.path.exists(self._model_directory):
            shutil.rmtree(self._model_directory)

    def _print_logs(self):
        """Prints logs from TensorRT conversion"""
        for chunk in self._logs:
            for line in chunk.decode().split("\n"):
                if "[V]" in line:
                    trt_logger.debug(line)
                elif "[W]" in line:
                    trt_logger.warning(line)
                elif "[E]" in line:
                    trt_logger.error(line)
                else:
                    trt_logger.info(line)

    def get_status(self) -> Optional[int]:
        """
        Get exit status of TensorRT conversion to check if the process was successful.

        Returns:
            int:
                None if the process is still running,
                0 if the process exited successfully,
                any other exit code means an error has occurred.
        """
        return self._api_client.exec_inspect(self._exec_id).get("ExitCode")



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   commands.py
@Author  :   Wei Loon Cheng, Kai Xuan Lee
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Custom runner installation commands.
"""

PIP_INSTALL_PACKAGE_COMMAND = "{} -m pip install {}"
MICROK8S_CONFIG_COMMAND = "microk8s kubectl config view --raw"
MICROK8S_STATUS_COMMAND = "microk8s status --wait-ready"



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   functions.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature CLI Batch Functions.
"""

import logging
import os
import sys
from datetime import datetime

from colorama import Fore
from InquirerPy import inquirer
from InquirerPy.base.control import Choice

from datature.nexus import error
from datature.nexus.api.batch import models
from datature.nexus.cli import consts
from datature.nexus.cli.functions import get_default_datature_client
from datature.nexus.utils import utils

logging.basicConfig(level=logging.CRITICAL)


def select_entry(entity: str) -> str:
    """
    Selects an entry from the list of entities.

    :param entity: The entity to select from.
    :return: The selected entry ID.
    """
    sdk_project = get_default_datature_client()
    batch_entity = getattr(sdk_project.batch, entity)

    try:
        response = batch_entity.list()
        response.data.reverse()
    except error.Error as exc:
        print(f"{Fore.CYAN}●{Fore.RESET} Error listing {entity}: {exc.message}")
        sys.exit(1)

    if len(response.data) == 0:
        print(f"{Fore.CYAN}●{Fore.RESET} No {entity} found.")
        sys.exit(0)

    if entity == "datasets":
        choices = [
            Choice(
                name=(
                    f"{item.name}  [Expires in "
                    + utils.timedelta_to_str(
                        datetime.fromtimestamp(int(item.expiry_time / 1000))
                        - datetime.now()
                    )
                    + "]"
                ),
                value=item.id,
            )
            for item in response.data
        ]
    else:
        choices = [Choice(name=item.name, value=item.id) for item in response.data]

    return inquirer.fuzzy(
        message="Which entry do you want to select?",
        choices=choices,
        max_height=consts.INQUIRER_CHOICES_MAX_VISIBLE_ENTRIES,
        border=True,
    ).execute()


def clear_terminal_screen():
    """Clear the terminal screen."""
    os.system("cls" if os.name == "nt" else "clear")


def clear_last_lines(num_lines=1):
    """Clear the last n lines of the terminal."""
    for _ in range(num_lines):
        # Move the cursor up one line
        sys.stdout.write("\033[F")
        # Clear the entire line
        sys.stdout.write("\033[K")


def handle_paged_navigation(response, pagination) -> bool:
    """Handle paged navigation for a paginated response."""
    page_choices = (
        [
            (Choice(name="Next Page", value="next") if response.next_page else None),
            (
                Choice(name="Previous Page", value="prev")
                if response.prev_page
                else None
            ),
            Choice(name="Exit", value="exit"),
        ]
        if response.next_page or response.prev_page
        else []
    )
    page_choices = [choice for choice in page_choices if choice]

    if not page_choices:
        return False

    selection = inquirer.select(
        message="Select an action:", choices=page_choices, border=True
    ).execute()

    if selection == "next":
        pagination["page"] = response.next_page
    elif selection == "prev":
        pagination["page"] = response.prev_page
    elif selection == "exit":
        return False

    clear_last_lines(len(response.data) + 3)
    return True


def print_webhook_object(webhook: models.WebhookModel, header_width: int = 20):
    """
    Print the webhook object.

    :param webhook: The webhook object.
    :param header_width: The width of the header.
    """
    print()  # for new line
    print(f"{'ID:':<{header_width}} {webhook.id}")
    print(
        f"{'Name:':<{header_width}} {utils.truncate_text(webhook.name, max_length=70)}"
    )
    print(f"{'Endpoint:':<{header_width}} {webhook.endpoint}")
    print(f"{'Max Retries:':<{header_width}} {webhook.retries.max_retries}")
    print(
        f"{'Max Retry Delay:':<{header_width}} {round(webhook.retries.max_retry_delay / 1000, 2)}s"
    )
    print()  # for new line


def print_dataset_object(dataset: models.Dataset, header_width: int = 15):
    """
    Print the dataset object.

    :param dataset: The dataset object.
    :param header_width: The width of the header.
    """
    print()  # for new line
    print(f"{'ID:':<{header_width}} {dataset.id}")
    print(
        f"{'Name:':<{header_width}} {utils.truncate_text(dataset.name, max_length=70)}"
    )
    print(f"{'Source:':<{header_width}} {dataset.status.source.kind}")
    print(
        f"{'Expires At:':<{header_width}} "
        f"{datetime.fromtimestamp(int(dataset.expiry_time / 1000))}"
    )
    print()  # for new line
    print(f"{'Status:':<{header_width}} {dataset.status.overview}")
    print(f"{'Message:':<{header_width}} {dataset.status.message}")
    print(f"{'Item Count:':<{header_width}} {dataset.status.item_count}")
    print(
        f"{'Last Updated:':<{header_width}} "
        f"{datetime.fromtimestamp(int(dataset.status.update_time / 1000))}"
    )
    print()  # for new line


def print_job_object(job: models.Job, dataset_count: int, header_width: int = 25):
    """
    Print the job object.

    :param job: The job object.
    :param dataset_count: The total number of items in the dataset.
    :param header_width: The width of the header.
    """
    print()  # for new line
    print(f"{'ID:':<{header_width}} {job.id}")
    print(f"{'Name:':<{header_width}} {utils.truncate_text(job.name, max_length=70)}")
    print(
        f"{'Start Time:':<{header_width}} "
        f"{datetime.fromtimestamp(int(job.spec.start_at_time / 1000))}"
    )
    print(
        f"{'Cutoff Time:':<{header_width}} "
        f"{datetime.fromtimestamp(int(job.spec.stop_at_time / 1000))}"
    )
    print(f"{'Dataset ID:':<{header_width}} {job.spec.dataset_id}")
    print()  # for new line
    print("Endpoint:")
    print(
        f"{'  - Webhook:':<{header_width}} {job.spec.result_delivery.destinations[0].webhook_id}"
    )
    print()  # for new line
    print(f"{'Status:':<{header_width}} {job.status.overview}")
    print(f"{'Message:':<{header_width}} {job.status.message}")
    print(f"{'Reason:':<{header_width}} {job.status.reason}")

    print()  # for new line
    print(f"{'Items Processed':<{header_width}} {'Count':<10} {'Description'}")

    print(
        f"{'  - Total:':<{header_width}} {dataset_count:<10} "
        f"{'Total item count in the Dataset.':<40}"
    )
    print(
        f"{'  - Gathered:':<{header_width}} {job.status.items.gathered:<10} "
        f"{job.status.items.get_description('gathered')}"
    )
    print(
        f"{'  - FailedGather:':<{header_width}} {job.status.items.failed_gather:<10} "
        f"{job.status.items.get_description('failed_gather')}"
    )
    print(
        f"{'  - Preprocessed:':<{header_width}} {job.status.items.preprocessed:<10} "
        f"{job.status.items.get_description('preprocessed')}"
    )
    print(
        f"{'  - FailedPreprocess:':<{header_width}} {job.status.items.failed_preprocess:<10} "
        f"{job.status.items.get_description('failed_preprocess')}"
    )
    print(
        f"{'  - Predicted:':<{header_width}} {job.status.items.predicted:<10} "
        f"{job.status.items.get_description('predicted')}"
    )
    print(
        f"{'  - FailedPredict:':<{header_width}} {job.status.items.failed_predict:<10} "
        f"{job.status.items.get_description('failed_predict')}"
    )
    print(
        f"{'  - Delivered:':<{header_width}} {job.status.items.delivered:<10} "
        f"{job.status.items.get_description('delivered')}"
    )
    print(
        f"{'  - FailedDeliver:':<{header_width}} {job.status.items.failed_deliver:<10} "
        f"{job.status.items.get_description('failed_deliver')}"
    )
    print()  # for new line



#!/usr/env/bin python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   tensorrt.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   TensorRT main module.
"""

# pylint: disable=E1102,R0902,R0913,R0914,R1732,W0511

import subprocess
import time
from contextlib import contextmanager
from typing import Generator, Optional

import requests

try:
    import docker
except ImportError as import_exc:
    raise ModuleNotFoundError(
        "Docker not installed! "
        "Please visit https://developers.datature.io/docs/tensorrt-export "
        "to learn how to set up your environment."
    ) from import_exc

from alive_progress import alive_bar

from datature.nexus import config, error
from datature.nexus.api.project import Project

from .converter import Converter
from .logger import trt_logger
from .params import ConversionParams, DockerParams
from .predictor import (
    MaskPredictor,
    ObjectDetectionPredictor,
    SemanticPredictor,
    VideoClassificationPredictor,
    YOLOv8BoxPredictor,
    YOLOv8ClassificationPredictor,
    YOLOv8MaskPredictor,
    YOLOv8PosePredictor,
)


class TensorRTConverter:
    """The TensorRT class provides an interface to the TensorRT Converter and Predictor."""

    _predictor_classes = {
        "od": ObjectDetectionPredictor,
        "mask": MaskPredictor,
        "semantic": SemanticPredictor,
        "video-cls": VideoClassificationPredictor,
        "yolov8": YOLOv8BoxPredictor,
        "yolov8-seg": YOLOv8MaskPredictor,
        "yolov8-pose": YOLOv8PosePredictor,
        "yolov8-cls": YOLOv8ClassificationPredictor,
        "yolov9": YOLOv8BoxPredictor,
    }

    def __init__(self, docker_params: Optional[DockerParams] = DockerParams()):
        """
        Instantiates the TensorRT Converter.

        Args:
            docker_params (DockerParams):
                Docker parameters for conversion and inference.
        """
        self._docker_params = docker_params
        self._converter = None
        self._inference_server = None
        self._docker_client = docker.from_env()
        self._api_client = docker.APIClient()

    def convert(
        self,
        project: Optional[Project] = None,
        artifact_id: Optional[str] = "",
        model_path: Optional[str] = "",
        output_path: Optional[str] = "",
        conversion_params: ConversionParams = ConversionParams(),
    ) -> None:
        """
        Instantiates the TensorRT Converter and runs the conversion process.

        Args:
            project (Project): The Nexus project instance containing the model to be converted.
            artifact_id (str): The artifact ID of the model to be converted.
            model_path (str): The path to the saved model directory.
            output_path (str): The path to the output directory.
            conversion_params (ConversionParams): Dictionary of conversion parameters.

        Raises:
            ValueError: If neither a local path to an ONNX model nor a Project instance is provided.
        """
        try:
            if not model_path and not project:
                raise ValueError(
                    "Either a local path to an ONNX model or a Project instance "
                    "needs to be provided if you wish to export an ONNX model "
                    "from your project on Datature Nexus"
                )
            trt_logger.info(
                "TensorRT conversion process may take up to 30 minutes, please be patient..."
            )
            with alive_bar(
                4 if project else 2,
                title="Running TensorRT Conversion",
                title_length=30,
                stats=False,
            ) as progress_bar:
                progress_bar.text("Setting up environment and resources")
                self._converter = Converter(
                    docker_client=self._docker_client,
                    api_client=self._api_client,
                    docker_params=self._docker_params,
                    conversion_params=conversion_params,
                )
                self._converter.setup()
                progress_bar()

                if project:
                    progress_bar.text("Exporting TensorRT-compatible ONNX model")
                    model_id = self._converter.create_export(
                        project=project, artifact_id=artifact_id
                    )
                    progress_bar.text(
                        "TensorRT-compatible ONNX model model successfully exported!"
                    )
                    progress_bar()
                    progress_bar.text("Downloading exported model")
                    model_path = self._converter.download_export(
                        project=project, model_id=model_id
                    )
                    progress_bar.text(
                        "TensorRT-compatible ONNX model model successfully downloaded!"
                    )
                    progress_bar()

                progress_bar.text("Converting to TensorRT")
                self._converter.convert(
                    model_path=model_path,
                    output_path=output_path,
                )
                progress_bar()
        except Exception as exc:  # pylint: disable=W0703
            trt_logger.error(exc)

    @contextmanager
    def inference_server(self, model_folder: str, timeout: int = 180) -> Generator:
        """
        Context manager to start and stop the NVIDIA Triton Inference Server Docker container.

        Args:
            model_folder (str): The path to the model folder containing the model files
                to be loaded into the inference server.
            timeout (int):
                The maximum time in seconds to wait for the inference server to be ready,
                defaults to 180.

        Raises:
            Error: If the inference server fails to start.
        """
        try:
            self._start_inference_server(model_folder, timeout)
            yield
        finally:
            self._stop_inference_server()

    def predict(
        self,
        task: str,
        input_path: str,
        model_name: str,
        label_map_path: str,
        skeleton_file_path: str = "",
        threshold: float = 0.7,
        save: bool = True,
        output_path: str = "",
        dtype: str = "float32",
    ) -> None:
        """
        Instantiates the TensorRT Predictor and runs the prediction process.

        Args:
            task (str):
                The task to be performed, either "od", "mask", "semantic", "yolov8",
                "yolov8-seg", "yolov8-pose", "yolov8-cls", or "yolov9".
            input_path (str): The path to the input image or video file.
            model_name (str): The name of the model to be used for prediction.
            label_map_path (str): The path to the label map file.
            skeleton_file_path (str): The path to the skeleton file for pose estimation.
            threshold (float): The confidence threshold for object detection, defaults to 0.7.
            save (bool): Whether to save the output, defaults to True.
            output_path (str): The path to the output directory, defaults to "".
            dtype (str):
                The data type of the model, either "float32" or "float16", defaults to "float32".
        """
        kwargs = {}
        if task == "yolov8-pose" and skeleton_file_path != "":
            kwargs["skeleton_file_path"] = skeleton_file_path

        try:
            predictor = self._predictor_classes.get(task)(
                dtype=dtype,
                input_path=input_path,
                model_name=model_name,
                label_map_path=label_map_path,
                threshold=threshold,
                save=save,
                output_path=output_path,
                **kwargs,
            )
        except TypeError as type_exc:
            trt_logger.error(type_exc, exc_info=True)
            trt_logger.error(
                "Prediction task not supported, valid tasks are one of "
                "['od', 'mask', 'semantic', 'yolov8', 'yolov8-seg', "
                "'yolov8-pose', 'yolov8-cls', 'yolov9']. "
                "Please refer to https://developers.datature.io/docs/tensorrt-export "
                "for more information on how to define the inference commands."
            )
            return

        try:
            predictor()
        except KeyboardInterrupt:
            trt_logger.info("Keyboard interrupt detected")
        except Exception as exc:  # pylint: disable=W0703
            trt_logger.error(exc, exc_info=True)

    def get_status(self):
        """Retrieve TensorRT conversion status."""
        self._converter.get_status()

    def _start_inference_server(self, model_folder: str, timeout: int = 180) -> None:
        """
        Starts the NVIDIA Triton Inference Server Docker container.

        Args:
            model_folder (str): The path to the model folder containing the model files
                to be loaded into the inference server.
            timeout (int):
                The maximum time in seconds to wait for the inference server to be ready,
                defaults to 180.

        Raises:
            Error: If the inference server fails to start.
        """
        try:
            candidate = self._docker_client.images.list(
                filters={"reference": self._docker_params.inference_docker_image}
            )

            if len(candidate) == 0:
                trt_logger.info(
                    "Pulling Docker image %s",
                    self._docker_params.inference_docker_image,
                )
                self._docker_client.images.pull(
                    self._docker_params.inference_docker_image
                )
        except docker.errors.DockerException as docker_exc:
            raise ConnectionError(
                f"Could not not pull Docker image {self._docker_params.inference_docker_image}"
            ) from docker_exc

        command = [
            "docker",
            "run",
            "--gpus",
            "all",
            "--rm",
            "-p",
            "8000:8000",
            "-p",
            "8001:8001",
            "-p",
            "8002:8002",
            "-v",
            f"{model_folder}:/models",
            self._docker_params.inference_docker_image,
            "tritonserver",
            "--model-repository=/models",
        ]

        trt_logger.info("Starting inference server...")
        try:
            self._inference_server = subprocess.Popen(
                command, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
            )
        except subprocess.CalledProcessError as sp_exc:
            raise ConnectionError(sp_exc) from sp_exc

        time.sleep(5)
        start = time.time()
        while int(time.time() - start) < timeout:
            try:
                response = requests.get(
                    "http://localhost:8000/v2/health/ready",
                    timeout=config.REQUEST_TIME_OUT_SECONDS,
                )

                if response.status_code == 200:
                    trt_logger.info("Inference server ready!")
                    return

                raise error.InternalServerError(
                    "Something went wrong with starting the Triton Inference Server, "
                    "please visit https://developers.datature.io/docs/tensorrt-export "
                    "for more information on how to set up the inference server."
                )
            except requests.exceptions.ConnectionError as conn_exc:
                trt_logger.error(conn_exc)
                trt_logger.warning(
                    "Trying to establish connection to inference server..."
                )
                time.sleep(10)

        raise ConnectionRefusedError(
            "Could not connect to the Triton Inference Server, "
            "please visit https://developers.datature.io/docs/tensorrt-export "
            "for more information on how to set up the inference server."
        )

    def _stop_inference_server(self) -> None:
        """Stops the NVIDIA Triton Inference Server Docker container."""
        if self._inference_server is None:
            trt_logger.warning(
                "Inference server is not either not running or could not be found."
            )
            trt_logger.warning(
                "Please run `docker ps -a` to check if the inference server is already running."
            )
            return

        if self._inference_server.poll() is None:
            trt_logger.info("Stopping inference server...")
            self._inference_server.terminate()



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   nii_processor.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Nii Processor
"""

import shutil
import tempfile
from os import makedirs, path
from pathlib import Path
from typing import List

import cv2
import matplotlib.image as mpImage
import nibabel as nib
import numpy as np

from datature.nexus import error

from .base_processor import BaseProcessor


class NiiProcessor(BaseProcessor):
    """Nii processor class"""

    def valid(self, request_data):
        """Valid the input

        :param request_data: The request data, include file path.
        :return: None.
        """
        file_path = request_data.get("file")

        if not file_path:
            raise error.BadRequestError("Required field, must include file path")

    def process(self, request_data) -> List[str]:
        """Start process file to asset video

        :param request_data: The request data, include file path.
        :return: str: The generate video path.
        """
        out_path = tempfile.mkdtemp()

        file_path = request_data.get("file")
        orientation = request_data.get("options", {}).get("nifti_orientation", None)

        file_name = Path(file_path).stem

        video_output = path.join(out_path, file_name)

        if not path.exists(video_output):
            makedirs(video_output)

        scan = nib.load(file_path)

        # Read data and get scan's shape
        scan_array = scan.get_fdata()

        # Get image pixel
        scan_headers = scan.header
        pix_dim = scan_headers["pixdim"][1:4]

        # Calculate new image dimensions from aspect ratio
        new_scan_dims = np.multiply(scan_array.shape, pix_dim)
        new_scan_dims = (
            round(new_scan_dims[0]),
            round(new_scan_dims[1]),
            round(new_scan_dims[2]),
        )

        # if client provide the orientation
        if orientation and orientation in ["x", "y", "z"]:
            output_file_path = f"{video_output}/{file_name}-{orientation}.mp4"

            self.__write_video(scan_array, new_scan_dims, orientation, output_file_path)

            return [output_file_path]

        # if not then save all orientations
        output_file_paths = []
        for orientation in ["x", "y", "z"]:
            output_file_path = f"{video_output}/{file_name}-{orientation}.mp4"

            self.__write_video(scan_array, new_scan_dims, orientation, output_file_path)

            output_file_paths.append(output_file_path)
        return output_file_paths

    def __write_video(self, scan_array, scan_dims, orientation, saved_path):
        """Write video to file

        :param scan_array: The NIfTI scan array data.
        :param scan_dims: The image dimensions of the file.
        :param orientation: The orientation of the pics.
        :param saved_path: The saved file path.
        """
        axis = {"x": 0, "y": 1, "z": 2}[orientation]
        size = {
            "x": [scan_dims[2], scan_dims[1]],
            "y": [scan_dims[2], scan_dims[0]],
            "z": [scan_dims[1], scan_dims[0]],
        }[orientation]

        four_cc = cv2.VideoWriter_fourcc(*"mp4v")
        video_writer = cv2.VideoWriter(saved_path, four_cc, 30.0, size)

        cache_path = tempfile.mkdtemp()

        for i in range(scan_array.shape[axis]):
            # Get slice along the correct axis and resize
            slice_axis = scan_array.take(i, axis=axis)

            # Convert to Grayscale and write to video file
            mpImage.imsave(
                f"{cache_path}/{orientation}-{str(i)}.png", slice_axis, cmap="gray"
            )
            new_image = cv2.imread(f"{cache_path}/{orientation}-{str(i)}.png")

            new_image = cv2.resize(new_image, size)
            video_writer.write(new_image)
        shutil.rmtree(cache_path)
        video_writer.release()



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   error_fixture.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Error Response Test Data
"""

forbidden_error_response = {
    "code": "ForbiddenError",
    "message": "No Access Rights to Requested Resources",
}

not_found_error_response = {
    "code": "ResourceNotFoundError",
    "message": "Requested Project Resource Not Found",
}

too_many_requests_error_response = {
    "code": "TooManyRequestsError",
    "message": "Too Many Requests",
}

internal_server_error_response = {
    "code": "InternalServerError",
    "message": "Internal Server Error",
}



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   __init__.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Processor package
"""

import pathlib
from os import PathLike

from .dicom_processor import DicomProcessor
from .nii_processor import NiiProcessor


def get_processor(file_path: PathLike):
    """Get processor"""

    extension = pathlib.Path(file_path).suffix.upper()

    if extension == ".DCM":
        return DicomProcessor()
    if extension == ".NII":
        return NiiProcessor()

    raise NotImplementedError



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   ontology.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Ontology API
"""
# pylint: disable=R0903

from datature.nexus import models
from datature.nexus.client_context import RestContext


class Ontology(RestContext):
    """Datature ontology API Resource."""

    def list_schemas(self) -> models.Ontologies:
        """Lists all training runs regardless of status.

        :return: A msgspec struct containing
            the ontology metadata with the following structure:

            .. code-block:: json

                [Ontology(
                    id='ontology_843e486c-58d7-45a7-b722-f4948e204a56',
                    project_id='proj_cd067221d5a6e4007ccbb4afb5966535',
                    index=1,
                    name='person',
                    color='#c1ff72',
                    description='TEST'
                    attributes=[OntologyAttribute(
                      id='attri_d0877827-63d6-4794-b520-ef3e0c57ef71',
                      name='Tag Group',
                      description='',
                      type='Categorical',
                      required=True,
                      options=OntologyAttributeOptions(
                        categories=['person', 'dog']
                      ),
                      default=['person']
                    )]
                )]

        :example:
            .. code-block:: python

            from datature.nexus import Client

            project = Client("5aa41e8ba........").get_project("proj_b705a........")

            ontologies = project.ontologies.list_schemas()
        """
        return self.requester.GET(
            f"/projects/{self.project_id}/ontologies", response_type=models.Ontologies
        )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   test_project.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Project API Test Cases
"""

import unittest
from unittest.mock import ANY, MagicMock

from datature.nexus.api.project import Project
from datature.nexus.client_context import ClientContext

# pylint: disable=W0212


class TestProject(unittest.TestCase):
    """Datature Project API Resource Test Cases."""

    def __init__(self, *args, **kwargs):
        """init global variables."""
        unittest.TestCase.__init__(self, *args, **kwargs)

        context = ClientContext("secret_key", endpoint="http://localhost:8080")
        context.project_id = "project_id"
        context.requester._request = MagicMock()

        self.project = Project(context)

    def test_retrieve(self):
        """Test retrieve a project."""
        self.project.get_info()

        self.project._context.requester._request.assert_called_once_with(
            "GET", "/projects/project_id", ANY, query=None, request_params=None
        )

    def test_modify(self):
        """Test update a project."""

        self.project.update({"name": "New Project Name"})

        self.project._context.requester._request.assert_called_once_with(
            "PATCH",
            "/projects/project_id",
            ANY,
            request_body={"name": "New Project Name"},
            query=None,
            request_params=None,
        )

    def test_insight(self):
        """Test retrieve a project insight."""
        self.project.list_insights()

        self.project._context.requester._request.assert_called_once_with(
            "GET", "/projects/project_id/insights", ANY, query=None, request_params=None
        )

    def test_users(self):
        """Test retrieve project users."""
        self.project.list_users()

        self.project._context.requester._request.assert_called_once_with(
            "GET", "/projects/project_id/users", ANY, query=None, request_params=None
        )



# !/usr/env/bin python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   types.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Custom model config types module.
"""
# pylint: disable=C0203, E1133

from enum import Enum, EnumMeta, unique


class FormatMeta(EnumMeta):
    """Metaclass for format options."""

    def __contains__(self, item):
        return item in [member.value for member in self]

    def __str__(self) -> str:
        return str([format.value for format in self])

    def __values__(self) -> list:
        return [format.value for format in self]


@unique
class ConfigRequiredFields(Enum, metaclass=FormatMeta):
    """Config fields for custom models."""

    TASK = "task"
    INPUTS = "inputs"
    OUTPUTS = "outputs"


@unique
class TaskTypes(Enum, metaclass=FormatMeta):
    """Task types for custom models."""

    CLASSIFICATION = "classification"
    BBOX = "bbox"
    SEMANTIC = "semantic"


class InputRequiredFields(Enum, metaclass=FormatMeta):
    """Input required fields for custom models."""

    CLASSIFICATION = ["image"]
    BBOX = ["image"]
    SEMANTIC = ["image"]


@unique
class OutputRequiredFields(Enum, metaclass=FormatMeta):
    """Output required fields for custom models."""

    CLASSIFICATION = ["class_scores"]
    BBOX = ["class_scores", "bounding_boxes"]
    SEMANTIC = ["mask_scores"]


@unique
class InputFormatOptions(Enum, metaclass=FormatMeta):
    """Input format options for custom models."""

    NCHW = ["n", "c", "h", "w"]
    NHWC = ["n", "h", "w", "c"]


@unique
class ClassificationOutputFormatOptions(Enum, metaclass=FormatMeta):
    """Output format options for image classification models."""

    CLASS_SCORES = [["batch_size", "num_classes"]]


class BboxFormats(Enum, metaclass=FormatMeta):
    """Output format options for bounding boxes."""

    XYXY = ["batch_size", "num_det", ["xmin", "ymin", "xmax", "ymax"]]
    YXYX = ["batch_size", "num_det", ["ymin", "xmin", "ymax", "xmax"]]
    XYWH = ["batch_size", "num_det", ["xmin", "ymin", "width", "height"]]
    CXCYWH = ["batch_size", "num_det", ["cx", "cy", "width", "height"]]


@unique
class BboxOutputFormatOptions(Enum, metaclass=FormatMeta):
    """Output format options for custom models."""

    CLASS_SCORES = [["batch_size", "num_det", "num_classes"]]
    BOUNDING_BOXES = BboxFormats.__values__()
    BOX_SCORES = [["batch_size", "num_det"]]


class SemanticOutputFormatOptions(Enum, metaclass=FormatMeta):
    """Output format options for semantic segmentation models."""

    MASK_SCORES = [["batch_size", "num_classes", "height", "width"]]


class OutputFormatOptions(Enum, metaclass=FormatMeta):
    """Output format options for custom models."""

    CLASSIFICATION = ClassificationOutputFormatOptions
    BBOX = BboxOutputFormatOptions
    SEMANTIC = SemanticOutputFormatOptions



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   utils.py
@Author  :   Wei Loon Cheng, Kai Xuan Lee
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Custom runner utility functions.
"""

# pylint: disable=W0719,E1102

import base64
import hashlib
import json
import os
import queue
import shlex
import shutil
import subprocess
import sys
import threading
import time
from functools import wraps
from pathlib import Path
from typing import List
from uuid import UUID

import requests
from alive_progress import alive_bar
from colorama import Fore
from halo import Halo
from wcwidth import wcswidth

from datature.nexus.cli.runner import commands as COMMANDS
from datature.nexus.cli.runner import consts as CONSTS
from datature.nexus.cli.runner import k8s_config as K8S_CONFIG
from datature.nexus.cli.runner import messages as MESSAGES
from datature.nexus.cli.runner.common import install_pip_package
from datature.nexus.cli.runner.config import (
    GPUInfo,
    RunnerConfig,
    RunnerMetadata,
    SystemSpecs,
)
from datature.nexus.config import REQUEST_TIME_OUT_SECONDS
from datature.nexus.error import Error

try:
    import kubernetes as k8s
except ImportError as exc:
    if not install_pip_package("kubernetes"):
        raise Error(
            MESSAGES.INSTALL_PIP_PACKAGE_ERROR_MESSAGE.format(package_name="kubernetes")
        ) from exc
    import kubernetes as k8s

try:
    import pynvml
except ImportError as exc:
    if not install_pip_package("nvidia-ml-py"):
        raise Error(
            MESSAGES.INSTALL_PIP_PACKAGE_ERROR_MESSAGE.format(
                package_name="nvidia-ml-py"
            )
        ) from exc
    import pynvml


def start_spinner(func: callable = None) -> callable:
    """Start spinner for the function."""

    def decorator(func: callable):

        @wraps(func)
        def wrapper(*args, **kwargs):
            wait_spinner = Halo(spinner="dots")
            message_queue = queue.Queue()
            stop_thread = threading.Event()

            def update_spinner():
                while not stop_thread.is_set():
                    try:
                        new_message = message_queue.get(timeout=0.1)
                        if not kwargs.get("quiet", False) and new_message is not None:
                            wait_spinner.text = new_message
                    except queue.Empty:
                        continue

            thread = threading.Thread(target=update_spinner)
            wait_spinner.start()
            thread.start()

            try:
                result = func(*args, **kwargs, message_queue=message_queue)
                if not kwargs.get("quiet", False):
                    wait_spinner.succeed(result)
                return result
            except Error as exc:
                error_symbol = f"{Fore.CYAN}●{Fore.RESET}"
                wait_spinner.stop_and_persist(error_symbol, exc.message)
                sys.exit(1)
            except Exception as exc:
                error_symbol = f"{Fore.CYAN}●{Fore.RESET}"
                wait_spinner.stop_and_persist(error_symbol, str(exc))
                raise Exception(exc) from exc
            finally:
                stop_thread.set()
                thread.join()
                wait_spinner.stop()

        return wrapper

    if func is None:
        return decorator
    return decorator(func)


@start_spinner
def check_system_specs(runner_config: RunnerConfig, message_queue: queue.Queue) -> None:
    """Check system specifications.

    :param runner_config (RunnerConfig): Runner configuration.
    :param message_queue (queue.Queue): Message queue.
    """
    try:
        num_cores = int(
            subprocess.check_output(["nproc", "--all"], stderr=subprocess.STDOUT)
            .strip()
            .decode()
        )
        ram_total = int(
            subprocess.check_output(["free", "-b"], stderr=subprocess.STDOUT)
            .split()[7]
            .decode()
        )
        if ram_total < CONSTS.MINIMUM_RAM_BYTES:
            raise Error(
                MESSAGES.INSUFFICIENT_RAM_ERROR_MESSAGE.format(
                    ram_total // 1e9, CONSTS.MINIMUM_RAM_BYTES
                )
            )
        free_space = shutil.disk_usage(Path("/")).free
        if free_space < CONSTS.MINIMUM_STORAGE_BYTES:
            raise Error(
                MESSAGES.INSUFFICIENT_STORAGE_ERROR_MESSAGE.format(
                    free_space // 1e6, CONSTS.MINIMUM_STORAGE_BYTES
                )
            )
        runner_config.system_specs = SystemSpecs(
            num_cores=num_cores, ram_total=ram_total, free_space=free_space
        )
    except subprocess.CalledProcessError as exc:
        raise Error(
            MESSAGES.CHECK_SYSTEM_SPECS_ERROR_MESSAGE.format(exc.stderr)
        ) from exc

    try:
        check_nvidia_drivers(runner_config, message_queue)
    except Error as exc:
        raise exc from exc

    return MESSAGES.CHECK_SYSTEM_SPECS_SUCCESS_MESSAGE


def check_nvidia_drivers(
    runner_config: RunnerConfig, message_queue: queue.Queue
) -> None:
    """Check NVIDIA drivers.

    :param runner_config (RunnerConfig): Runner configuration.
    :param message_queue (queue.Queue): Message queue.
    """
    try:
        message_queue.put(MESSAGES.CHECK_DRIVERS_MESSAGE)
        pynvml.nvmlInit()
        driver_version = pynvml.nvmlSystemGetDriverVersion()
        if isinstance(driver_version, bytes):
            driver_version = driver_version.decode("utf-8")
        driver_version_major = int(driver_version.split(".")[0])
        if driver_version_major < 515 or driver_version_major > 535:
            raise Error(MESSAGES.DRIVER_INCOMPATIBLE_ERROR_MESSAGE)
        cuda_version = pynvml.nvmlSystemGetCudaDriverVersion()

        message_queue.put(MESSAGES.FETCH_GPU_INFO_MESSAGE)
        device_count = pynvml.nvmlDeviceGetCount()
        if device_count == 0:
            raise Error(MESSAGES.NO_GPU_FOUND_ERROR_MESSAGE)

        if len(runner_config.gpus) == 0:
            for i in range(device_count):
                handle = pynvml.nvmlDeviceGetHandleByIndex(i)
                device_name = str(pynvml.nvmlDeviceGetName(handle))
                device_vram_bytes = int(pynvml.nvmlDeviceGetMemoryInfo(handle).total)
                device_compute_capability = pynvml.nvmlDeviceGetCudaComputeCapability(
                    handle
                )
                device_uuid = pynvml.nvmlDeviceGetUUID(handle)
                if isinstance(device_uuid, bytes):
                    device_uuid = device_uuid.decode("utf-8")
                num_cuda_cores = pynvml.nvmlDeviceGetNumGpuCores(handle)
                runner_config.gpus.append(
                    GPUInfo(
                        device_id=i,
                        device_name=device_name,
                        device_vram_bytes=device_vram_bytes,
                        device_compute_capability=int(
                            "".join(map(str, device_compute_capability))
                        ),
                        device_driver_version=driver_version,
                        device_cuda_version=cuda_version,
                        device_uuid=device_uuid.replace("GPU-", ""),
                        device_cuda_cores=num_cuda_cores,
                    )
                )
        pynvml.nvmlShutdown()
    except subprocess.CalledProcessError as exc:
        raise Error(f"Error running nvidia-settings: {exc.stderr}") from exc
    except pynvml.NVMLError as exc:
        raise Error(MESSAGES.NVML_ERROR_MESSAGE.format(str(exc))) from exc
    except Error as exc:
        raise exc from exc


@start_spinner
def configure_microk8s(message_queue: queue.Queue) -> None:
    """Configure MicroK8s.

    :param message_queue (queue.Queue): Message queue.
    """
    try:
        message_queue.put(MESSAGES.CHECK_MICROK8S_STATUS_MESSAGE)
        subprocess.check_output(
            shlex.split(COMMANDS.MICROK8S_STATUS_COMMAND), stderr=subprocess.STDOUT
        )
    except FileNotFoundError as exc:
        raise Error(MESSAGES.MICROK8S_NOT_FOUND_ERROR_MESSAGE) from exc
    except subprocess.CalledProcessError as exc:
        raise Error(MESSAGES.CHECK_MICROK8S_STATUS_ERROR_MESSAGE) from exc
    return MESSAGES.CHECK_MICROK8S_STATUS_SUCCESS_MESSAGE


def ping_server(runner_config: RunnerConfig) -> None:
    """Ping Datature server.

    :param runner_config (RunnerConfig): Runner configuration
    """
    runner_config.status.valid_secret_key = False
    try:
        # Dummy ping to Datature server to check for internet connection.
        response = requests.get(
            f"{CONSTS.NEXUS_ENDPOINT}/workspaces/ws_{runner_config.workspace_id}/runners",
            headers={
                "Secret-Key": runner_config.secret_key,
                "Content-Type": "application/json",
            },
            timeout=REQUEST_TIME_OUT_SECONDS,
        )
        response.raise_for_status()
    except requests.exceptions.ConnectionError as exc:
        raise Error(MESSAGES.PING_SERVER_CONNECTION_ERROR_MESSAGE) from exc
    except requests.exceptions.HTTPError as exc:
        if response.status_code == 402:
            print(MESSAGES.FEATURE_NOT_AVAILABLE_ERROR_MESSAGE)
            sys.exit(1)
        if response.status_code == 403:
            raise Error(
                MESSAGES.PING_SERVER_FORBIDDEN_ERROR_MESSAGE.format(
                    runner_config.workspace_id
                )
            ) from exc
        if response.status_code != 405:
            raise Error(
                MESSAGES.PING_SERVER_UNKNOWN_ERROR_MESSAGE.format(response.text)
            ) from exc
    runner_config.status.valid_secret_key = True


def create_runner_id(runner_config: RunnerConfig) -> str:
    """Create runner UUID based on machine UUID and GPU UUIDs.

    :param runner_config (RunnerConfig): Runner configuration.
    :return (str): Runner UUID.
    """
    with open("/etc/machine-id", "r", encoding="utf-8") as machine_id_file:
        machine_id = str(machine_id_file.read().strip())
    runner_id_string = machine_id + runner_config.workspace_id
    for gpu in runner_config.gpus:
        runner_id_string += gpu.device_uuid

    sha256_hash = hashlib.sha256(runner_id_string.encode()).hexdigest()
    hex_uuid = (
        sha256_hash[:8]
        + "-"
        + sha256_hash[8:12]
        + "-"
        + "4"
        + sha256_hash[13:16]
        + "-"
        + sha256_hash[16:20]
        + "-"
        + sha256_hash[20:32]
    )
    return f"runner_{str(UUID(hex_uuid))}"


@start_spinner
def register_runner(runner_config: RunnerConfig, message_queue: queue.Queue) -> None:
    """Register runner with Datature server.

    :param runner_config (RunnerConfig): Runner configuration.
    :param message_queue (queue.Queue): Message queue.
    """
    message_queue.put(MESSAGES.REGISTER_RUNNER_MESSAGE.format(runner_config.name))
    try:
        response = requests.post(
            f"{CONSTS.NEXUS_ENDPOINT}/workspaces/ws_{runner_config.workspace_id}/runners",
            headers={
                "Secret-Key": runner_config.secret_key,
                "Content-Type": "application/json",
            },
            data=json.dumps(K8S_CONFIG.get_resource_config(runner_config)),
            timeout=REQUEST_TIME_OUT_SECONDS,
        )
        response.raise_for_status()
    except requests.exceptions.ConnectionError as exc:
        raise Error(
            MESSAGES.REGISTER_RUNNER_CONNECTION_ERROR_MESSAGE.format(runner_config.name)
        ) from exc
    except requests.exceptions.HTTPError as exc:
        if "No access rights" in response.text:
            raise Error(
                MESSAGES.REGISTER_RUNNER_FORBIDDEN_ERROR_MESSAGE.format(
                    runner_config.name, runner_config.workspace_id
                )
            ) from exc
        if "Conflict" in response.text:
            raise Error(
                MESSAGES.REGISTER_RUNNER_CONFLICT_ERROR_MESSAGE.format(
                    runner_config.name
                )
            ) from exc
        raise Error(
            MESSAGES.REGISTER_RUNNER_UNKNOWN_ERROR_MESSAGE.format(
                response.text, runner_config.name
            )
        ) from exc
    except requests.exceptions.RequestException as exc:
        raise Error(
            MESSAGES.REGISTER_RUNNER_UNKNOWN_ERROR_MESSAGE.format(
                response.text, runner_config.name
            )
        ) from exc

    runner_config.kubectl_namespace = "datature-" + runner_config.id.replace(
        "runner_", "runner-"
    )

    return MESSAGES.REGISTER_RUNNER_SUCCESS_MESSAGE.format(
        runner_config.name, runner_config.id[-6:]
    )


@start_spinner
def create_runner_namespace(
    runner_config: RunnerConfig,
    k8s_core_api_instance: k8s.client.CoreV1Api,
    message_queue: queue.Queue,
) -> None:
    """Create runner namespace in Kubernetes.

    :param runner_config (RunnerConfig): Runner configuration.
    :param k8s_core_api_instance (k8s.client.CoreV1Api): Kubernetes Core API instance.
    :param message_queue (queue.Queue): Message queue.
    """
    message_queue.put(
        MESSAGES.CREATE_RUNNER_NAMESPACE_MESSAGE.format(runner_config.kubectl_namespace)
    )

    try:
        runner_namespace = k8s.client.V1Namespace(
            api_version="v1",
            kind="Namespace",
            metadata={"name": runner_config.kubectl_namespace},
        )
        k8s_core_api_instance.create_namespace(body=runner_namespace)
    except k8s.client.rest.ApiException as exc:
        if (
            exc.status != 409
            and json.loads(exc.body).get("reason", "") != "AlreadyExists"
        ):
            raise Error(
                MESSAGES.CREATE_RUNNER_NAMESPACE_ERROR_MESSAGE.format(
                    exc.body, runner_config.kubectl_namespace
                )
            ) from exc
    return MESSAGES.CREATE_RUNNER_NAMESPACE_SUCCESS_MESSAGE.format(
        runner_config.kubectl_namespace
    )


def get_runner_metadata(runner_config: RunnerConfig) -> None:
    """Get runner metadata.

    :param runner_config (RunnerConfig): Runner configuration.
    """
    try:
        response = requests.get(
            f"{CONSTS.NEXUS_ENDPOINT}/workspaces/ws_{runner_config.workspace_id}/metadata/runner",
            headers={
                "Secret-Key": runner_config.secret_key,
                "Content-Type": "application/json",
            },
            timeout=REQUEST_TIME_OUT_SECONDS,
        )
        runner_config.runner_metadata = RunnerMetadata.from_dict(response.json())
    except requests.exceptions.ConnectionError as exc:
        raise Error(MESSAGES.PING_SERVER_CONNECTION_ERROR_MESSAGE) from exc
    except requests.exceptions.HTTPError as exc:
        if response.status_code == 402:
            print(MESSAGES.FEATURE_NOT_AVAILABLE_ERROR_MESSAGE)
            sys.exit(1)
        if response.status_code == 403:
            raise Error(
                MESSAGES.PING_SERVER_FORBIDDEN_ERROR_MESSAGE.format(
                    runner_config.workspace_id
                )
            ) from exc
        if response.status_code != 405:
            raise Error(
                MESSAGES.PING_SERVER_UNKNOWN_ERROR_MESSAGE.format(response.text)
            ) from exc


@start_spinner
def apply_config_map(
    runner_config: RunnerConfig,
    k8s_core_api_instance: k8s.client.CoreV1Api,
    message_queue: queue.Queue,
) -> None:
    """Apply configuration map for runner in Kubernetes.

    :param runner_config (RunnerConfig): Runner configuration.
    :param k8s_core_api_instance (k8s.client.CoreV1Api): Kubernetes Core API instance.
    :param message_queue (queue.Queue): Message queue.
    """
    message_queue.put(MESSAGES.APPLY_CONFIG_MAP_MESSAGE)

    is_existing_config_map = False
    try:
        k8s_core_api_instance.read_namespaced_config_map(
            name=CONSTS.RUNNER_DEPLOYMENT_CONFIG_NAME,
            namespace=runner_config.kubectl_namespace,
        )
        is_existing_config_map = True
    except k8s.client.rest.ApiException as exc:
        if exc.status != 404 and json.loads(exc.body).get("reason", "") != "NotFound":
            pass

    try:
        config_map = k8s.client.V1ConfigMap(
            api_version="v1",
            kind="ConfigMap",
            metadata={
                "name": CONSTS.RUNNER_DEPLOYMENT_CONFIG_NAME,
                "namespace": runner_config.kubectl_namespace,
            },
            data={
                "secret_key": runner_config.secret_key,
                "workspace_id": runner_config.workspace_id,
                "runner_id": runner_config.id,
                "runner_name": runner_config.name,
                "logs_directory": str(CONSTS.RUNNER_LOG_DIR),
                "docker_registry": os.getenv(
                    "DATATURE_RUNNER_DOCKER_REGISTRY",
                    runner_config.runner_metadata.runner_registry,
                ),
                "api_url": os.getenv(
                    "DATATURE_API_BASE_URL", "https://api.datature.io"
                ),
            },
        )

        if is_existing_config_map:
            k8s_core_api_instance.patch_namespaced_config_map(
                name=CONSTS.RUNNER_DEPLOYMENT_CONFIG_NAME,
                namespace=runner_config.kubectl_namespace,
                body=config_map,
            )
        else:
            k8s_core_api_instance.create_namespaced_config_map(
                namespace=runner_config.kubectl_namespace, body=config_map
            )
    except k8s.client.rest.ApiException as exc:
        if (
            exc.status != 409
            and json.loads(exc.body).get("reason", "") != "AlreadyExists"
        ):
            raise Error(
                MESSAGES.APPLY_CONFIG_MAP_ERROR_MESSAGE.format(
                    exc.body, runner_config.kubectl_namespace
                )
            ) from exc
    return MESSAGES.APPLY_CONFIG_MAP_SUCCESS_MESSAGE


@start_spinner
def apply_service_account(
    runner_config: RunnerConfig,
    k8s_core_api_instance: k8s.client.CoreV1Api,
    k8s_rbac_api_instance: k8s.client.RbacAuthorizationV1Api,
    message_queue: queue.Queue,
) -> None:
    """Apply service account for runner in Kubernetes.

    :param runner_config (RunnerConfig): Runner configuration.
    :param k8s_core_api_instance (k8s.client.CoreV1Api): Kubernetes Core API instance.
    :param k8s_rbac_api_instance (k8s.client.RbacAuthorizationV1Api): Kubernetes RBAC API instance.
    :param message_queue (queue.Queue): Message queue.
    """
    message_queue.put(MESSAGES.APPLY_SERVICE_ACCOUNT_MESSAGE)

    try:
        service_account = k8s.client.V1ServiceAccount(
            api_version="v1",
            kind="ServiceAccount",
            metadata={
                "name": "python-job-sa",
                "namespace": runner_config.kubectl_namespace,
            },
        )

        k8s_core_api_instance.create_namespaced_service_account(
            namespace=runner_config.kubectl_namespace, body=service_account
        )
        k8s_rbac_api_instance.create_namespaced_role(
            namespace=runner_config.kubectl_namespace,
            body=K8S_CONFIG.get_role(runner_config.kubectl_namespace),
        )
        k8s_rbac_api_instance.create_namespaced_role_binding(
            namespace=runner_config.kubectl_namespace,
            body=K8S_CONFIG.get_role_binding(runner_config.kubectl_namespace),
        )
    except k8s.client.rest.ApiException as exc:
        if (
            exc.status != 409
            and json.loads(exc.body).get("reason", "") != "AlreadyExists"
        ):
            raise Error(
                MESSAGES.APPLY_SERVICE_ACCOUNT_ERROR_MESSAGE.format(
                    exc.body, runner_config.kubectl_namespace
                )
            ) from exc
    return MESSAGES.APPLY_SERVICE_ACCOUNT_SUCCESS_MESSAGE


@start_spinner
def setup_docker_registry_credentials(
    runner_config: RunnerConfig,
    k8s_core_api_instance: k8s.client.CoreV1Api,
    message_queue: queue.Queue,
) -> None:
    """Setup Docker registry credentials for runner in Kubernetes.

    :param runner_config (RunnerConfig): Runner configuration.
    :param k8s_core_api_instance (k8s.client.CoreV1Api): Kubernetes Core API instance.
    :param message_queue (queue.Queue): Message queue.
    """
    message_queue.put(
        MESSAGES.DELETE_DOCKER_REGCRED_MESSAGE.format(CONSTS.DOCKER_REGISTRY_SECRET)
    )
    try:
        # Delete existing secret datature-regcred if it exists
        body = k8s.client.V1DeleteOptions()
        k8s_core_api_instance.delete_namespaced_secret(
            name=CONSTS.DOCKER_REGISTRY_SECRET,
            namespace=runner_config.kubectl_namespace,
            body=body,
            grace_period_seconds=0,
        )
    except k8s.client.rest.ApiException as exc:
        if exc.status != 404 and json.loads(exc.body).get("reason", "") != "NotFound":
            raise Error(
                MESSAGES.DELETE_DOCKER_REGCRED_ERROR_MESSAGE.format(
                    CONSTS.DOCKER_REGISTRY_SECRET
                )
            ) from exc

    # Create new docker-registry secret datature-regcred
    message_queue.put(
        MESSAGES.CREATE_DOCKER_REGCRED_MESSAGE.format(CONSTS.DOCKER_REGISTRY_SECRET)
    )

    try:
        docker_config_dict = {
            "auths": {
                os.getenv(
                    "DATATURE_RUNNER_DOCKER_REGISTRY",
                    runner_config.runner_metadata.runner_registry,
                ): {
                    "username": "secret-key",
                    "password": runner_config.secret_key,
                }
            }
        }
        docker_config = base64.b64encode(
            json.dumps(docker_config_dict).encode()
        ).decode()
        body = k8s.client.V1Secret(
            api_version="v1",
            kind="Secret",
            type="kubernetes.io/dockerconfigjson",
            data={".dockerconfigjson": docker_config},
            metadata={
                "name": CONSTS.DOCKER_REGISTRY_SECRET,
                "namespace": runner_config.kubectl_namespace,
            },
        )
        k8s_core_api_instance.create_namespaced_secret(
            namespace=runner_config.kubectl_namespace, body=body
        )
    except k8s.client.rest.ApiException as exc:
        raise Error(
            MESSAGES.CREATE_DOCKER_REGCRED_ERROR_MESSAGE.format(
                CONSTS.DOCKER_REGISTRY_SECRET
            )
        ) from exc
    return MESSAGES.SETUP_DOCKER_REGCRED_SUCCESS_MESSAGE


@start_spinner
def apply_deployment(
    runner_config: RunnerConfig,
    k8s_apps_api_instance: k8s.client.AppsV1Api,
    message_queue: queue.Queue,
) -> None:
    """Apply deployment for runner in Kubernetes.

    :param runner_config (RunnerConfig): Runner configuration.
    :param k8s_apps_api_instance (k8s.client.AppsV1Api): Kubernetes Apps API instance.
    :param message_queue (queue.Queue): Message queue.
    """
    message_queue.put(
        MESSAGES.APPLY_DEPLOYMENT_MESSAGE.format(
            runner_config.name, runner_config.id[-6:]
        )
    )
    logs_folder = CONSTS.RUNNER_LOG_DIR / runner_config.id

    try:
        deployment = k8s.client.V1Deployment(
            api_version="apps/v1",
            kind="Deployment",
            metadata={
                "name": CONSTS.RUNNER_DEPLOYMENT_NAME,
                "namespace": runner_config.kubectl_namespace,
                "labels": {"app": CONSTS.RUNNER_DEPLOYMENT_NAME},
            },
            spec=K8S_CONFIG.get_deployment_spec(runner_config, logs_folder),
        )
        k8s_apps_api_instance.create_namespaced_deployment(
            namespace=runner_config.kubectl_namespace, body=deployment
        )
    except k8s.client.rest.ApiException as exc:
        if (
            exc.status != 409
            and json.loads(exc.body).get("reason", "") != "AlreadyExists"
        ):
            raise Error(
                MESSAGES.APPLY_DEPLOYMENT_ERROR_MESSAGE.format(
                    exc.body,
                    CONSTS.RUNNER_DEPLOYMENT_NAME,
                    runner_config.kubectl_namespace,
                )
            ) from exc
    return MESSAGES.APPLY_DEPLOYMENT_SUCCESS_MESSAGE.format(
        CONSTS.RUNNER_DEPLOYMENT_NAME
    )


def initialize_runner(
    runner_config: RunnerConfig,
    k8s_apps_api_instance: k8s.client.AppsV1Api,
    k8s_custom_api_instance: k8s.client.CustomObjectsApi,
) -> None:
    """Initialize runner in Kubernetes.

    :param runner_config (RunnerConfig): Runner configuration.
    :param k8s_apps_api_instance (k8s.client.AppsV1Api): Kubernetes Apps API instance.
    :param k8s_custom_api_instance (k8s.client.CustomObjectsApi): Kubernetes Custom API instance.
    """
    try:
        patch = {
            "spec": {
                "validator": {
                    "driver": {
                        "env": [
                            {
                                "name": "DISABLE_DEV_CHAR_SYMLINK_CREATION",
                                "value": "true",
                            }
                        ]
                    }
                }
            }
        }
        k8s_custom_api_instance.patch_namespaced_custom_object(
            group="nvidia.com",
            version="v1",
            namespace="",
            plural="clusterpolicies",
            name="cluster-policy",
            body=patch,
        )
    except k8s.client.rest.ApiException as exc:
        if exc.status == 404:
            print(MESSAGES.PATCH_CLUSTER_POLICY_NOT_FOUND_ERROR_MESSAGE)
            raise Error(MESSAGES.PATCH_CLUSTER_POLICY_NOT_FOUND_ERROR_MESSAGE) from exc
        print(MESSAGES.PATCH_CLUSTER_POLICY_ERROR_MESSAGE.format(exc.body))
        raise Error(
            MESSAGES.PATCH_CLUSTER_POLICY_ERROR_MESSAGE.format(exc.body)
        ) from exc

    with alive_bar(
        CONSTS.RUNNER_INITIALIZATION_NUM_CHECKS,
        title=MESSAGES.INITIALIZE_RUNNER_MESSAGE,
        title_length=27,
        length=30,
        stats=True,
        elapsed=False,
    ) as progress_bar:
        for current_check in range(CONSTS.RUNNER_INITIALIZATION_NUM_CHECKS):
            response = k8s_apps_api_instance.read_namespaced_deployment(
                name=CONSTS.RUNNER_DEPLOYMENT_NAME,
                namespace=runner_config.kubectl_namespace,
            )
            if response.status.conditions is None:
                time.sleep(CONSTS.RUNNER_INITIALIZATION_SLEEP_SECONDS)
                progress_bar()
                continue

            runner_initialization_query = " ".join(
                [
                    f"{condition.status} {condition.type}"
                    for condition in response.status.conditions
                ]
            )

            if "True Available" in runner_initialization_query:
                progress_bar(CONSTS.RUNNER_INITIALIZATION_NUM_CHECKS - current_check)
                return

            time.sleep(CONSTS.RUNNER_INITIALIZATION_SLEEP_SECONDS)
            progress_bar()

    print(
        MESSAGES.INITIALIZE_RUNNER_ERROR_MESSAGE.format(
            runner_config.name, runner_config.id[-6:]
        )
    )
    raise Error(
        MESSAGES.INITIALIZE_RUNNER_ERROR_MESSAGE.format(
            runner_config.name, runner_config.id[-6:]
        )
    )


def truncate_text(text: str, max_length: int = 10) -> str:
    """Truncate text to a maximum length.

    :param text (str): Text to truncate.
    :param max_length (int): Maximum length, default is 10 characters.
    :return (str): Truncated text.
    """
    text_width = wcswidth(text)
    if text_width <= max_length:
        return text

    current_width = 0
    truncated_text = ""

    for char in text:
        char_width = wcswidth(char)
        if current_width + char_width > max_length - 3:
            break

        truncated_text += char
        current_width += char_width

    truncated_text += "..."
    space_padding = max_length - wcswidth(truncated_text)
    truncated_text += " " * space_padding
    return truncated_text


def format_runner_status(runner_config: RunnerConfig) -> None:
    """Format runner status and print to stdout.

    :param runner_config (RunnerConfig): Runner configuration.
    """
    name_padding = max(0, 26 - (wcswidth(runner_config.name) - len(runner_config.name)))
    full_status = runner_config.status.status + (
        " [SUSPENDED]" if runner_config.status.suspended else ""
    )
    ram_gib = str(int(runner_config.system_specs.ram_total / 1e9))
    cuda_major = runner_config.gpus[-1].device_cuda_version // 1e3
    cuda_minor = (runner_config.gpus[-1].device_cuda_version % 1e3) // 10
    cuda_version = f"{cuda_major}.{cuda_minor}"

    print(f"+{'-' * 73}+")
    print(f"| {'RUNNER NAME':<26} {'HASH':<10} {'STATUS':<22} {'SECRET KEY':<10} |")
    print(
        f"| {truncate_text(runner_config.name, 26):<{name_padding}} "
        f"{runner_config.id[-6:]:<10} {full_status:<22} "
        f"{'Valid' if runner_config.status.valid_secret_key else 'Invalid':<10} |"
    )
    print(f"|{'=' * 73}|")
    print(f"| {'Cores':<10} {'RAM':<15} {'GPUs':<10} {'CUDA':<10} {'Driver':<22} |")
    print(
        f"| {runner_config.system_specs.num_cores:<10} {ram_gib:<5} {'GiB':<9} "
        f"{len(runner_config.gpus):<10} {cuda_version:<10} "
        f"{runner_config.gpus[-1].device_driver_version:<22} |"
    )
    print(f"|{'-' * 73}|")
    print(f"| {'GPU':<10} {'Name':<26} {'VRAM':<10} {'CC':<22} |")
    for gpu in runner_config.gpus:
        gpu_device_name_padding = max(
            0, 26 - (wcswidth(gpu.device_name) - len(gpu.device_name))
        )
        vram_gib = str(int(gpu.device_vram_bytes / 1e9))
        print(
            f"| {gpu.device_id:<10} "
            f"{truncate_text(gpu.device_name, 26):<{gpu_device_name_padding}} "
            f"{vram_gib:<4} {'GiB':<5} {gpu.device_compute_capability:<22} |"
        )
    print(f"+{'-' * 73}+")

    if len(runner_config.status.runs) > 0:
        print(f"+{'-' * 80}+")
        print(
            f"| {'RUN HASH':<9} {'IMAGE':<15} {'STATUS':<12} "
            f"{'STARTED AT':<20} {'RAM':<10} {'GPUs':<7} |"
        )
        for run in runner_config.status.runs:
            run_hash = run.id[-6:]
            run_allocated_memory_gib = str(int(run.allocated_memory / 1e9))
            run_image_name_padding = max(
                0, 15 - (wcswidth(run.image_name) - len(run.image_name))
            )

            print(
                f"| {run_hash:<9} {truncate_text(run.image_name, 15):<{run_image_name_padding}} "
                f"{run.status:<12} {run.started:<20} "
                f"{run_allocated_memory_gib:<4} {'GiB':<5} {run.num_gpus:<7} |"
            )
            print(f"+{'-' * 80}+")


def format_runner_list(runner_configs: List[RunnerConfig]) -> None:
    """Format runner list and print to stdout.

    :param runner_configs (List[RunnerConfig]): List of runner configurations.
    """
    print(f"{'RUNNER NAME':<26} {'HASH':<10} {'STATUS':<22} {'SECRET KEY':<10}")

    for runner_config in runner_configs:
        name_padding = max(
            0, 26 - (wcswidth(runner_config.name) - len(runner_config.name))
        )
        full_status = runner_config.status.status + (
            " [SUSPENDED]" if runner_config.status.suspended else ""
        )
        runner_hash = runner_config.id[-6:]
        print(
            f"{truncate_text(runner_config.name, 26):<{name_padding}} "
            f"{runner_hash:<10} {full_status:<22} "
            f"{'Valid' if runner_config.status.valid_secret_key else 'Invalid':<10}"
        )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   __init__.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   SDK init module
"""

from .api import types as ApiTypes
from .client import Client
from .version import __version__

# Expose certain elements at package level
__all__ = ["Client", "ApiTypes", "__version__"]



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   asset_fixture.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Assets Test Data
"""

# pylint: disable=C0301
list_assets_response = {
    "next_page": "ZjY0YTdiYTI0NzAzNThlMDljZGNjYzY4Zg",
    "data": [
        {
            "id": "asset_e478b4d2-496b-4abe-a1d2-3cccad605002",
            "object": "asset",
            "filename": "boat180.png",
            "project": "proj_306c543ddd4ff1188a6e4b43a586b62a",
            "create_date": 1688713764134,
            "metadata": {
                "file_size": 186497,
                "mime_type": "image/png",
                "status": "annotated",
                "height": 243,
                "width": 400,
                "groups": ["main"],
                "custom_metadata": {},
            },
            "statistic": {
                "tags_count": [{"name": "boat", "count": 1}],
                "total_annotations": 1,
            },
            "url": "https://storage.googleapis.com/assets.datature.io/306c543ddd4ff1188a6e4b43a586b62a/assets/boat180.png?X-Goog-Algorithm=GOOG4-RSA-SHA256&X-Goog-Credential=gcs-asset-signer%40datature-nexus.iam.gserviceaccount.com%2F20230707%2Fauto%2Fstorage%2Fgoog4_request&X-Goog-Date=20230707T071051Z&X-Goog-Expires=86400&X-Goog-SignedHeaders=host&generation=1688713773057960&X-Goog-Signature=81ed17f20a5af973c6c4a95e3ea454751f83abc91c0b95999329f68b85724f51f9df032789a7e29bb4bc09b6acffbbd4fca56b2e1c7ca133dff29c79088bbfcd123117ad31aacf7e146de6f98c90a9958985a1d38e781c4b2afae3a15c011404229d29445634b92f71468b78b50aec20bc8a10781305fac529d0c6b9b444c91536d95b2490acdc19341a2126aac40b1013514ef4b2082075c047bbcae1ffeb74044c6ec093d2dab6c391f4f3a886613c9faeb78e06ce4ccf7b428651623e4afc38a544fc4c8d934a9dbeb4ca61f68c103252e43a2248461613758e3421a4737707177ef16983e97aedd6aa36adfdd963c1b7a1d19c81abcdff0efd9373056c59",
        },
    ],
}

annotations_response = [
    {
        "id": "annot_1e1a0c12-cf6a-42cb-8347-6596006621b5",
        "object": "annotation",
        "bound_type": "rectangle",
        "bound": [
            [0.425, 0.49382716049382713],
            [0.425, 0.6419753086419753],
            [0.6, 0.6419753086419753],
            [0.6, 0.49382716049382713],
        ],
        "project_id": "proj_306c543ddd4ff1188a6e4b43a586b62a",
        "asset_id": "asset_e478b4d2-496b-4abe-a1d2-3cccad605002",
        "tag": "boat",
    }
]



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   setup.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Setup module
"""

import re

import setuptools

# read the contents of your README file
with open("README.md", "r", encoding="utf8") as rd:
    long_description = rd.read()

# read the version number
with open("datature/nexus/version.py", "r", encoding="utf8") as rd:
    version = re.search(r'__version__\s*=\s*[\'"]([^\'"]*)[\'"]', rd.read()).group(1)

REQUIRES = []
with open("requirements.txt", "r", encoding="utf8") as f:
    for line in f:
        line, _, _ = line.partition("#")
        line = line.strip()
        if not line or line.startswith("setuptools"):
            continue
        REQUIRES.append(line)

setuptools.setup(
    name="datature",
    version=version,
    author="Datature",
    license="Apache License 2.0",
    author_email="developers@datature.io",
    long_description_content_type="text/markdown",
    long_description=long_description,
    url="https://datature.io/",
    project_urls={
        "Homepage": "https://datature.io/",
        "Documentation": "https://developers.datature.io/docs/python-sdk",
        "Download": "https://pypi.org/project/datature#files",
        "Changelog": "https://developers.datature.io/docs/sdk-changelog",
        "Slack Community": "https://datature.io/community",
    },
    description="Python bindings for the Datature API",
    packages=setuptools.find_namespace_packages(),
    python_requires=">=3.8",
    install_requires=REQUIRES,
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: Apache Software License",
    ],
    entry_points={
        "console_scripts": ["datature=datature.nexus.cli.main:main"],
    },
)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   test_client.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Project API Test Cases
"""

import json
import unittest
from unittest.mock import ANY, MagicMock

import msgspec

from datature.nexus import Client, models
from datature.nexus.api.project import Project
from tests.unit.fixture.data import projects_fixture

# pylint: disable=W0212


class TestClient(unittest.TestCase):
    """Datature Client Test Cases."""

    def __init__(self, *args, **kwargs):
        """init global variables."""
        unittest.TestCase.__init__(self, *args, **kwargs)

        client = Client("secret_key")
        client._context.requester._request = MagicMock()

        self.client = client

    def test_info(self):
        """Test retrieve workspace info."""
        self.client.get_info()

        self.client._context.requester._request.assert_called_once_with(
            "GET", "/workspace", ANY, query=None, request_params=None
        )

    def test_list_projects(self):
        """Test retrieve workspace projects."""
        self.client._context.requester._request.return_value = msgspec.json.decode(
            json.dumps(projects_fixture.list_projects_response),
            type=models.Projects,
        )

        self.client.list_projects()

    def test_get_project(self):
        """Test retrieve project ."""
        project = self.client.get_project("project")

        assert isinstance(project, Project)



#!/usr/env/bin python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   logger.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   TensorRT logging module.
"""

import logging

from colorama import Fore, Style

trt_logger = logging.getLogger("trt")
trt_logger.setLevel(logging.DEBUG)


class ColoredFormatter(logging.Formatter):
    """Logging colored formatter"""

    def __init__(self):
        super().__init__()
        self._formats = {
            logging.DEBUG: (
                f" {Fore.BLUE} %(levelname)s"
                f" {Fore.WHITE} %(message)s{Style.RESET_ALL}"
            ),
            logging.INFO: (
                f" {Style.RESET_ALL}{Fore.CYAN} %(levelname)s"
                f" {Fore.WHITE} %(message)s{Style.RESET_ALL}"
            ),
            logging.WARNING: (
                f" {Style.RESET_ALL}{Fore.YELLOW} %(levelname)s"
                f" {Fore.YELLOW + Style.BRIGHT} %(message)s{Style.RESET_ALL}"
            ),
            logging.ERROR: (
                f" {Style.RESET_ALL}{Fore.RED} %(levelname)s"
                f" {Style.BRIGHT} %(message)s{Style.RESET_ALL}"
            ),
            logging.CRITICAL: (
                f" {Style.RESET_ALL}{Fore.RED} %(levelname)s"
                f" {Style.BRIGHT} %(message)s{Style.RESET_ALL}"
            ),
        }

    def format(self, record):
        log_fmt = self._formats.get(record.levelno)
        date_fmt = "%Y-%m-%d %H:%M:%S"
        formatter = logging.Formatter(fmt=log_fmt, datefmt=date_fmt)
        return formatter.format(record)


stdout_handler = logging.StreamHandler()
stdout_handler.setLevel(logging.INFO)
stdout_handler.setFormatter(ColoredFormatter())

trt_logger.addHandler(stdout_handler)

warnings_logger = logging.getLogger("py.warnings")
warnings_logger.addHandler(stdout_handler)



#!/usr/env/bin python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   prediction_utils.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   TensorRT predictor util functions.
"""

# pylint: disable=R0912,R0913,R0914,R0915

import copy
import json
from typing import Dict, Tuple

import cv2
import numpy as np
import requests
from PIL import Image

# Triton Inference Server API endpoint
# default port is 8000 since Triton Inference Server uses 8000 as default
TRITON_MODEL_CONFIG_URL = "http://localhost:8000/v2/models/{}/config"


def get_model_config(model_name: str) -> Dict:
    """
    Get model configuration from Triton Inference Server.

    Args:
        model_name (str): Model name.

    Returns:
        model_config (dict): Model configuration.
    """
    model_config = requests.get(
        TRITON_MODEL_CONFIG_URL.format(model_name), timeout=(3, 10)
    ).json()
    return model_config


def load_label_map(label_map_path: str) -> Dict:
    """
    Reads label map in the format of .pbtxt and parse into dictionary

    Args:
      label_map_path: File path to the label_map

    Returns:
      Dictionary with the format of {label_index: {'id': label_index, 'name': label_name}}
    """
    label_map = {}

    with open(label_map_path, "r", encoding="utf-8") as label_file:
        for line in label_file:
            if "id" in line:
                label_index = int(line.split(":")[-1])
                label_name = next(label_file).split(":")[-1].strip().strip('"')
                label_map[label_index] = {
                    "id": label_index,
                    "name": label_name,
                }

    return label_map


def load_skeleton(skeleton_path: str) -> Dict:
    """Load skeletons from JSON file.

    Args:
        skeleton_path (str): Path to skeleton JSON file.

    Returns:
        Dictionary of skeleton with name as key and connections by position.
    """

    def _keypoint_positions(skeleton):
        keypoint_name_list = [keypoint["name"] for keypoint in skeleton["keypoints"]]
        return {
            "keypoint_names": keypoint_name_list,
            "connections": [
                [
                    keypoint_name_list.index(con["pair"][0]),
                    keypoint_name_list.index(con["pair"][1]),
                ]
                for con in skeleton["connections"]
            ],
        }

    with open(skeleton_path, "r", encoding="utf-8") as skeleton_file:
        skeletons = json.load(skeleton_file)["skeletons"]
    skeleton_dict = {
        skeleton["name"]: _keypoint_positions(skeleton) for skeleton in skeletons
    }

    return skeleton_dict


def draw_bboxes(
    image: np.ndarray,
    box: np.ndarray,
    class_id: int,
    score: float,
    color_map: dict,
    category_index: dict,
) -> np.ndarray:
    """
    Draws bounding boxes and labels on the input image based on the detected objects.

    Args:
        image (np.ndarray): The input image to draw detections on.
        box (np.ndarray): Bounding box coordinates in the format of [ymin, xmin, ymax, xmax].
        class_id (int): Class id of the detected object.
        score (float): Confidence score of the detected object.
        color_map (dict): Dictionary containing color map for each class.
        category_index (dict): Dictionary containing class index and name.

    Returns:
        img (np.ndarray): The input image with drawn detections.
    """
    origi_shape = image.shape
    color = color_map.get(category_index[class_id]["id"]).tolist()
    cv2.rectangle(
        image,
        (
            int(box[1] * origi_shape[1]),  # x1
            int(box[0] * origi_shape[0]),  # y1
        ),
        (
            int(box[3] * origi_shape[1]),  # x2
            int(box[2] * origi_shape[0]),  # y2
        ),
        color,
        2,
    )
    cv2.rectangle(
        image,
        (
            int(box[1] * origi_shape[1]),
            int(box[2] * origi_shape[0]),
        ),
        (
            int(box[3] * origi_shape[1]),
            int(box[2] * origi_shape[0] + 15),
        ),
        color,
        -1,
    )
    cv2.putText(
        image,
        f"{str(category_index[class_id]['name'])}, {str(round(score, 2))}",
        (
            int(box[1] * origi_shape[1]),
            int(box[2] * origi_shape[0] + 10),
        ),
        cv2.FONT_HERSHEY_SIMPLEX,
        0.3,
        (0, 0, 0),
        1,
        cv2.LINE_AA,
    )
    return image


def draw_instance_masks(
    image: np.ndarray,
    box: np.ndarray,
    mask: np.ndarray,
    class_id: int,
    color_map: dict,
    category_index: dict,
):
    """Draw instance masks on image.

    Args:
        image (np.ndarray): Image to draw masks on.
        box (np.ndarray): Bounding box coordinates in the format of [ymin, xmin, ymax, xmax].
        mask (np.ndarray): Instance mask.
        class_id (int): Class id.
        color_map (dict): Dictionary containing color map for each class.
        category_index (dict): Dictionary containing class index and name.

    Returns:
        image (np.ndarray): Image with masks drawn on.
    """
    color = color_map.get(category_index[class_id]["id"]).tolist()

    mask = np.expand_dims(mask, -1)
    mask = np.tile(mask, 3)
    mask[..., :] *= np.array(color, dtype=np.uint8)
    mask = np.clip(mask, 0, 255).astype(np.uint8)
    xmin, ymin, xmax, ymax = box
    ymin = int(ymin * image.shape[1])
    ymax = int(ymax * image.shape[1])
    xmin = int(xmin * image.shape[0])
    xmax = int(xmax * image.shape[0])
    mask_height = ymax - ymin
    mask_width = xmax - xmin
    mask = cv2.resize(mask, (mask_height, mask_width))
    resized_mask = np.zeros(image.shape)
    resized_mask[xmin:xmax, ymin:ymax, :] = mask
    image = image.astype(np.int64)
    image += resized_mask.astype(np.int64)
    image = np.clip(image, 0, 255).astype(np.uint8)
    return image


def draw_yolov8_instance_masks(
    image: np.ndarray,
    box: np.ndarray,
    mask: np.ndarray,
    class_id: int,
    color_map: dict,
    category_index: dict,
):
    """Draw YOLOv8 instance masks on image.

    Args:
        image (np.ndarray): Image to draw masks on.
        box (np.ndarray): Bounding box coordinates in the format of [ymin, xmin, ymax, xmax].
        mask (np.ndarray): Instance mask.
        class_id (int): Class id.
        color_map (dict): Dictionary containing color map for each class.
        category_index (dict): Dictionary containing class index and name.

    Returns:
        image (np.ndarray): Image with masks drawn on.
    """
    color = color_map.get(category_index[class_id]["id"]).tolist()

    ymin, xmin, ymax, xmax = box
    mask_height, mask_width = mask.shape
    ymin = int(ymin * mask_height)
    xmin = int(xmin * mask_width)
    ymax = int(ymax * mask_height)
    xmax = int(xmax * mask_width)
    row = np.expand_dims(np.arange(mask_width), 0)
    col = np.expand_dims(np.arange(mask_height), -1)
    mask *= (row >= xmin) * (row <= xmax) * (col >= ymin) * (col <= ymax)
    mask = mask.astype(np.uint8)
    mask = np.expand_dims(mask, -1)
    mask = np.tile(mask, 3)
    mask = cv2.resize(
        mask, (image.shape[1], image.shape[0]), interpolation=cv2.INTER_LINEAR
    )
    mask = np.where(mask > 0, np.uint8(1), np.uint8(0))
    masked_image = image.copy()
    masked_image = np.where(mask, np.array(color, dtype=np.uint8), masked_image)
    image = cv2.addWeighted(image, 0.3, masked_image, 0.7, 0)
    return image


def draw_semantic_masks(image: np.ndarray, mask: np.ndarray, color_map: dict):
    """Draw semantic masks on image.

    Args:
        image (np.ndarray): Image to draw masks on.
        mask (np.ndarray): Semantic mask.

    Returns:
        image (np.ndarray): Image with masks drawn on.
    """
    mask = np.expand_dims(mask, -1)
    mask = np.tile(mask, 3)
    unique = np.unique(mask)
    output_mask = np.zeros_like(mask, dtype=np.uint8)
    for item in unique:
        output_mask = np.where(mask != item, output_mask, color_map[int(item)])
    output_mask = cv2.resize(output_mask, (image.shape[1], image.shape[0]))
    image = image.astype(np.uint8)
    pil_image = Image.fromarray(image)
    pil_mask = Image.fromarray(output_mask)
    pil_image = Image.blend(pil_image, pil_mask, 0.7)
    return np.array(pil_image)


def draw_keypoints(
    image: np.ndarray,
    keypoints,
    scores,
    class_id,
    color_map: dict,
    category_index: dict,
    skeletons: dict,
):
    """Draw keypoints and skeleton on image.

    Args:
        image (np.ndarray): Image to draw keypoints on.
        keypoints (np.ndarray): Array of keypoints.
        scores (np.ndarray): Array of keypoint confidences.
        class_id (int): Class id.
        color_map (dict): Dictionary containing color map for each class.
        category_index (dict): Dictionary containing class index and name.
        skeletons (dict): Dictionary containing skeleton information.

    Returns:
        image (np.ndarray): Image with keypoints and skeleton drawn on.
    """
    height, width, _ = image.shape
    _, ndim = keypoints.shape
    skeleton = skeletons.get(category_index[int(class_id)]["name"])
    if skeleton is None:
        return image

    for ske in skeleton["connections"]:
        pos1 = (
            int(keypoints[(int(ske[0])), 0] * width),
            int(keypoints[(int(ske[0])), 1] * height),
        )
        pos2 = (
            int(keypoints[(int(ske[1])), 0] * width),
            int(keypoints[(int(ske[1])), 1] * height),
        )
        if ndim == 3:
            conf1 = keypoints[(ske[0]), 2]
            conf2 = keypoints[(ske[1]), 2]
            if conf1 < 0.5 or conf2 < 0.5:
                continue
        if pos1[0] % width == 0 or pos1[1] % height == 0 or pos1[0] < 0 or pos1[1] < 0:
            continue
        if pos2[0] % width == 0 or pos2[1] % height == 0 or pos2[0] < 0 or pos2[1] < 0:
            continue
        cv2.line(
            image,
            pos1,
            pos2,
            color_map[class_id].tolist(),
            thickness=2,
            lineType=cv2.LINE_AA,
        )
    for kpi, (each_kp, each_score) in enumerate(zip(keypoints, scores)):
        x_coord, y_coord = each_kp[0] * width, each_kp[1] * height
        if x_coord % width != 0 and y_coord % height != 0:
            if len(each_kp) == 3:
                conf = each_kp[2]
                if conf < 0.5:
                    continue
            cv2.circle(
                image,
                (int(x_coord), int(y_coord)),
                2,
                (0, 0, 0),
                -1,
                lineType=cv2.LINE_AA,
            )
            cv2.rectangle(
                image,
                (int(x_coord) - 10, int(y_coord) - 15),
                (
                    int(x_coord)
                    + len(
                        f"{skeleton['keypoint_names'][kpi]} "
                        f"{str(round(each_score, 2))}"
                    )
                    * 5,
                    int(y_coord),
                ),
                (0, 0, 0),
                -1,
            )
            cv2.putText(
                image,
                f"{skeleton['keypoint_names'][kpi]} " f"{str(round(each_score, 2))}",
                (int(x_coord) - 10, int(y_coord) - 5),
                cv2.FONT_HERSHEY_SIMPLEX,
                0.3,
                (255, 255, 255),
                1,
                cv2.LINE_AA,
            )
    return image


def draw_class_labels(
    image: np.ndarray,
    class_id: int,
    score: float,
    color_map: dict,
    category_index: dict,
):
    """Draw class labels on image.

    Args:
        image (np.ndarray): Image to draw class labels on.
        class_id (int): Class id.
        score (float): Confidence score.
        color_map (dict): Dictionary containing color map for each class.
        category_index (dict): Dictionary containing class index and name.

    Returns:
        image (np.ndarray): Image with class labels drawn on.
    """
    height, width, _ = image.shape
    color = color_map.get(class_id - 1).tolist()
    cv2.rectangle(
        image,
        (0, 0),
        (width, height),
        color,
        4,
    )
    cv2.rectangle(
        image,
        (0, 0),
        (width, 15),
        color,
        -1,
    )
    cv2.putText(
        image,
        f"{str(category_index[class_id]['name'])}, {str(round(score, 2))}",
        (10, 10),
        cv2.FONT_HERSHEY_SIMPLEX,
        0.3,
        (0, 0, 0),
        1,
        cv2.LINE_AA,
    )
    return image


def postprocess_predictions(pred: np.array, num_classes: int) -> Tuple[np.ndarray]:
    """Postprocess predictions from model output.

    Args:
        pred (np.array): Model output.
        num_classes (int): Number of classes.

    Returns:
        scores (np.ndarray): Array of scores.
        classes (np.ndarray): Array of classes.
        boxes_xywh (np.ndarray): Array of bounding boxes in xywh format.
        extras (np.ndarray): Array of extra predictions.
    """
    num_extras = pred.shape[1] - num_classes - 4
    boxes_xywh = pred[..., :4]
    class_end = 4 + num_classes
    score_classes = pred[..., 4:class_end]
    extras = pred[..., class_end : class_end + num_extras]

    scores = np.max(score_classes, axis=1)
    classes = np.argmax(score_classes, axis=1).astype(np.int32)

    return scores, classes, boxes_xywh, extras


def xywh2xyxy(boxes_xywh: np.ndarray, height: int, width: int) -> np.ndarray:
    """Convert boxes from xywh to xyxy format.

    Args:
        boxes_xywh (np.ndarray): Array of bounding boxes in xywh format.
        height (int): Height of the image.
        width (int): Width of the image.

    Returns:
        boxes_xyxy (np.ndarray): Array of bounding boxes in xyxy format.
    """
    boxes_xyxy = np.copy(boxes_xywh)
    boxes_xyxy[..., 0] = (
        boxes_xywh[..., 0] - boxes_xywh[..., 2] / 2
    ) / width  # top left x
    boxes_xyxy[..., 1] = (
        boxes_xywh[..., 1] - boxes_xywh[..., 3] / 2
    ) / height  # top left y
    boxes_xyxy[..., 2] = (
        boxes_xywh[..., 0] + boxes_xywh[..., 2] / 2
    ) / width  # bottom right x
    boxes_xyxy[..., 3] = (
        boxes_xywh[..., 1] + boxes_xywh[..., 3] / 2
    ) / height  # bottom right y
    boxes_xyxy = np.clip(boxes_xyxy, 0, 1)
    return boxes_xyxy


def nms_boxes(
    boxes: np.ndarray,
    classes: np.ndarray,
    scores: np.ndarray,
    iou_threshold: float,
    dtype: str,
    masks: np.ndarray = None,
    keypoints: np.ndarray = None,
    keypoint_confs: np.ndarray = None,
    confidence: float = 0.5,
    sigma: float = 0.5,
) -> Tuple[np.ndarray]:
    """Carry out non-max suppression on the detected bboxes.

    Args:
        boxes (np.ndarray): Array of bounding boxes in xyxy format.
        classes (np.ndarray): Array of classes.
        scores (np.ndarray): Array of scores.
        iou_threshold (float): IoU threshold for NMS.
        dtype (str): Data type of the model output.
        masks (np.ndarray): Array of masks.
        keypoints (np.ndarray): Array of keypoints.
        keypoint_confs (np.ndarray): Array of keypoint confidences.
        confidence (float): Confidence threshold for NMS.
        sigma (float): Sigma value for Soft-NMS.

    Returns:
        nboxes (np.ndarray): Array of non-max suppressed bounding boxes.
        nclasses (np.ndarray): Array of non-max suppressed classes.
        nscores (np.ndarray): Array of non-max suppressed scores.
        nmasks (np.ndarray): Array of non-max suppressed masks.
        nkpts (np.ndarray): Array of non-max suppressed keypoints.
        nkptcnfs (np.ndarray): Array of non-max suppressed keypoint confidences.
    """
    is_soft = False
    use_exp = False

    nboxes, nclasses, nscores, nmasks, nkpts, nkptcnfs = [], [], [], [], [], []
    for cls in set(classes):
        # handle data for one class
        inds = np.where(classes == cls)
        bbx = boxes[inds]
        cls = classes[inds]
        sco = scores[inds]
        if masks is not None:
            msk = masks[inds]
        if keypoints is not None and keypoint_confs is not None:
            kpt = keypoints[inds]
            kptcnf = keypoint_confs[inds]

        # make a data copy to avoid breaking
        # during nms operation
        b_nms = copy.deepcopy(bbx)
        c_nms = copy.deepcopy(cls)
        s_nms = copy.deepcopy(sco)
        if masks is not None:
            m_nms = copy.deepcopy(msk)
        if keypoints is not None and keypoint_confs is not None:
            k_nms = copy.deepcopy(kpt)
            kc_nms = copy.deepcopy(kptcnf)

        while len(s_nms) > 0:
            # pick the max box and store, here
            # we also use copy to persist result
            i = np.argmax(s_nms, axis=-1)
            nboxes.append(copy.deepcopy(b_nms[i]))
            nclasses.append(copy.deepcopy(c_nms[i]))
            nscores.append(copy.deepcopy(s_nms[i]))
            if masks is not None:
                nmasks.append(copy.deepcopy(m_nms[i]))
            if keypoints is not None and keypoint_confs is not None:
                nkpts.append(copy.deepcopy(k_nms[i]))
                nkptcnfs.append(copy.deepcopy(kc_nms[i]))

            # swap the max line and first line
            b_nms[[i, 0], :] = b_nms[[0, i], :]
            c_nms[[i, 0]] = c_nms[[0, i]]
            s_nms[[i, 0]] = s_nms[[0, i]]
            if masks is not None:
                m_nms[[i, 0]] = m_nms[[0, i]]
            if keypoints is not None and keypoint_confs is not None:
                k_nms[[i, 0], :] = k_nms[[0, i], :]
                kc_nms[[i, 0]] = kc_nms[[0, i]]

            iou = box_diou(b_nms, dtype)

            # drop the last line since it has been record
            b_nms = b_nms[1:]
            c_nms = c_nms[1:]
            s_nms = s_nms[1:]
            if masks is not None:
                m_nms = m_nms[1:]
            if keypoints is not None and keypoint_confs is not None:
                k_nms = k_nms[1:]
                kc_nms = kc_nms[1:]

            if is_soft:
                # Soft-NMS
                if use_exp:
                    # score refresh formula:
                    # score = score * exp(-(iou^2)/sigma)
                    s_nms = s_nms * np.exp(-(iou * iou) / sigma)
                else:
                    # score refresh formula:
                    # score = score * (1 - iou) if iou > threshold
                    depress_mask = np.where(iou > iou_threshold)[0]
                    s_nms[depress_mask] = s_nms[depress_mask] * (1 - iou[depress_mask])
                keep_mask = np.where(s_nms >= confidence)[0]
            else:
                # normal Hard-NMS
                keep_mask = np.where(iou <= iou_threshold)[0]

            # keep needed box for next loop
            b_nms = b_nms[keep_mask]
            c_nms = c_nms[keep_mask]
            s_nms = s_nms[keep_mask]
            if masks is not None:
                m_nms = m_nms[keep_mask]
            if keypoints is not None and keypoint_confs is not None:
                k_nms = k_nms[keep_mask]
                kc_nms = kc_nms[keep_mask]

    # reformat result for output
    nboxes = np.array(nboxes)
    nclasses = np.array(nclasses)
    nscores = np.array(nscores)
    nmasks = np.array(nmasks)
    nkpts = np.array(nkpts)
    nkptcnfs = np.array(nkptcnfs)
    return nboxes, nclasses, nscores, nmasks, nkpts, nkptcnfs


def box_diou(boxes: np.ndarray, dtype: str) -> np.ndarray:
    """
    Calculate DIoU value of 1st box with other boxes of a box array
    Reference Paper:
        "Distance-IoU Loss: Faster and Better Learning for
        Bounding Box Regression"
        https://arxiv.org/abs/1911.08287

    Args:
        boxes (np.ndarray): Array of bounding boxes in xyxy format.
        dtype (str): Data type of the model output.

    Returns:
        diou (np.ndarray): Array of DIoU values.
    """
    # get box coordinate and area
    x_pos = boxes[:, 0]
    y_pos = boxes[:, 1]
    wid = boxes[:, 2]
    hei = boxes[:, 3]
    areas = wid * hei
    areas[areas == np.inf] = np.finfo(dtype).max  # account for possible overflow

    # check IoU
    inter_xmin = np.maximum(x_pos[1:], x_pos[0])
    inter_ymin = np.maximum(y_pos[1:], y_pos[0])
    inter_xmax = np.minimum(x_pos[1:] + wid[1:], x_pos[0] + wid[0])
    inter_ymax = np.minimum(y_pos[1:] + hei[1:], y_pos[0] + hei[0])

    inter_w = np.maximum(0.0, inter_xmax - inter_xmin + 1)
    inter_h = np.maximum(0.0, inter_ymax - inter_ymin + 1)

    inter = inter_w * inter_h
    iou = inter / (areas[1:] + areas[0] - inter)
    iou[iou == np.inf] = np.finfo(dtype).max  # account for possible overflow

    # box center distance
    x_center = x_pos + wid / 2
    y_center = y_pos + hei / 2
    center_distance = np.power(x_center[1:] - x_center[0], 2) + np.power(
        y_center[1:] - y_center[0], 2
    )
    center_distance[center_distance == np.inf] = np.finfo(
        dtype
    ).max  # account for possible overflow

    # get enclosed area
    enclose_xmin = np.minimum(x_pos[1:], x_pos[0])
    enclose_ymin = np.minimum(y_pos[1:], y_pos[0])
    enclose_xmax = np.maximum(x_pos[1:] + wid[1:], x_pos[0] + wid[0])
    enclose_ymax = np.maximum(y_pos[1:] + wid[1:], y_pos[0] + wid[0])
    enclose_w = np.maximum(0.0, enclose_xmax - enclose_xmin + 1)
    enclose_h = np.maximum(0.0, enclose_ymax - enclose_ymin + 1)
    # get enclosed diagonal distance
    enclose_diagonal = np.power(enclose_w, 2) + np.power(enclose_h, 2)
    enclose_diagonal[enclose_diagonal == np.inf] = np.finfo(
        dtype
    ).max  # account for possible overflow
    # calculate DIoU, add epsilon in denominator to avoid dividing by 0
    diou = iou - 1.0 * (center_distance) / (enclose_diagonal + np.finfo(float).eps)

    return diou


def clamp(val: float, minimum: float = 0.0, maximum: float = 1.0) -> float:
    """Clamp a value between a minimum and maximum value.

    Args:
        x (float): The value to be clamped.
        minimum (float): The minimum value, defaults to 0.0.
        maximum (float): The maximum value, defaults to 1.0.

    Returns:
        float: The clamped value.
    """
    return max(minimum, min(val, maximum))



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   types.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Types for Datature API resources.
"""
# pylint: disable=R0902,C0103,W0212

import logging
from dataclasses import dataclass
from enum import Enum, unique
from typing import List, Optional, Union

logger = logging.getLogger("datature-nexus")


def add_mapping(enum_cls):
    """Add Mapping to Type."""

    enum_cls._member_map_.update(enum_cls.__MAPPING__)
    return enum_cls


@dataclass
class ProjectMetadata:
    """Project metadata.

    :param name: The name of the project.
    """

    name: str

    def to_json(self):
        """Function to convert dataclass to dict"""
        return {
            "name": self.name,
        }


@add_mapping
@unique
class BoundType(Enum):
    """Bound Type.

    Bound Type Options:
        Rectangle
        Polygon
        Classification
        Keypoints
    """

    POLYGON = "Polygon"
    RECTANGLE = "Rectangle"
    CLASSIFICATION = "Classification"
    KEYPOINTS = "Keypoints"

    __MAPPING__ = {
        "Polygon": POLYGON,
        "Rectangle": RECTANGLE,
        "Classification": CLASSIFICATION,
        "Keypoints": KEYPOINTS,
    }


@add_mapping
@unique
class AnnotationExportFormat(Enum):
    """Annotation Export CSV Format.

    Bounding Box Options:
        coco
        csv_fourcorner
        csv_widthheight
        pascal_voc
        yolo_darknet
        yolo_keras_pytorch
        createml
        tfrecord

    Polygon Options:
        polygon_single
        polygon_coco

    Classification Options:
        csv_classification
        classification_tfrecord

    Key Point Options:
        keypoints_coco
    """

    COCO = "coco"
    CSV_FOURCORNER = "csv_fourcorner"
    CSV_WIDTHHEIGHT = "csv_widthheight"
    PASCAL_VOC = "pascal_voc"
    YOLO_DARKNET = "yolo_darknet"
    YOLO_KERAS_PYTORCH = "yolo_keras_pytorch"
    CREATEML = "createml"
    TFRECORD = "tfrecord"
    POLYGON_COCO = "polygon_coco"
    POLYGON_SINGLE = "polygon_single"
    CSV_CLASSIFICATION = "csv_classification"
    CLASSIFICATION_TFRECORD = "classification_tfrecord"
    KEYPOINTS_COCO = "keypoints_coco"

    __MAPPING__ = {
        "coco": COCO,
        "csv_fourcorner": CSV_FOURCORNER,
        "csv_widthheight": CSV_WIDTHHEIGHT,
        "pascal_voc": PASCAL_VOC,
        "yolo_darknet": YOLO_DARKNET,
        "yolo_keras_pytorch": YOLO_KERAS_PYTORCH,
        "createml": CREATEML,
        "tfrecord": TFRECORD,
        "polygon_coco": POLYGON_COCO,
        "polygon_single": POLYGON_SINGLE,
        "csv_classification": CSV_CLASSIFICATION,
        "classification_tfrecord": CLASSIFICATION_TFRECORD,
        "keypoints_coco": KEYPOINTS_COCO,
    }


@add_mapping
@unique
class AssetStatus(Enum):
    """Asset Status.

    Asset Status Options:
        Annotated
        Review
        Completed
        Tofix
        None
    """

    ANNOTATED = "Annotated"
    REVIEW = "Review"
    COMPLETED = "Completed"
    TOFIX = "Tofix"
    NONE = "None"

    __MAPPING__ = {
        "Annotated": ANNOTATED,
        "Review": REVIEW,
        "Completed": COMPLETED,
        "Tofix": TOFIX,
        "None": NONE,
    }


@add_mapping
@unique
class OperationStatusOverview(Enum):
    """Operation Status."""

    QUEUED = "Queued"
    RUNNING = "Running"
    FINISHED = "Finished"
    CANCELLED = "Cancelled"
    ERRORED = "Errored"

    __MAPPING__ = {
        "Queued": QUEUED,
        "Running": RUNNING,
        "Finished": FINISHED,
        "Cancelled": CANCELLED,
        "Errored": ERRORED,
    }


@add_mapping
@unique
class DeploymentStatusOverview(Enum):
    """Deployment Status."""

    CREATING = "Creating"
    UPDATING = "Updating"
    REMOVING = "Removing"
    AVAILABLE = "Available"
    ERRORED = "Errored"

    __MAPPING__ = {
        "Creating": CREATING,
        "Updating": UPDATING,
        "Removing": REMOVING,
        "Available": AVAILABLE,
        "Errored": ERRORED,
    }


@add_mapping
@unique
class RunStatusOverview(Enum):
    """Run Status."""

    CREATING = "Creating"
    RUNNING = "Running"
    FINISHED = "Finished"
    ERRORED = "Errored"
    CANCELLED = "Cancelled"

    __MAPPING__ = {
        "Creating": CREATING,
        "Running": RUNNING,
        "Finished": FINISHED,
        "Errored": ERRORED,
        "Cancelled": CANCELLED,
    }


@add_mapping
@unique
class ModelFormat(Enum):
    """Artifact supported Models Format."""

    TENSORFLOW = "TensorFlow"
    TFLITE = "TFLite"
    ONNX = "ONNX"
    PYTORCH = "PyTorch"
    COREML = "CoreML"

    __MAPPING__ = {
        "TensorFlow": TENSORFLOW,
        "TFLite": TFLITE,
        "ONNX": ONNX,
        "PyTorch": PYTORCH,
        "CoreML": COREML,
    }


DEFAULT_INSTANCES = [
    "instance_x1cpu-standard",
    "instance_t4-standard-1g",
    "instance_l4-standard-1g",
]


@dataclass
class AnnotationMetadata:
    """Annotation metadata.

    :param asset_id: The unique ID of the asset.
    :param tag: The tag class name of the annotation.
    :param bound_type: The bound type of the annotation (Rectangle or Polygon).
    :param bound: The bound coordinates of the annotation in
        [[x1, y1], [x2, y2], ... , [xn, yn]] format.
    """

    asset_id: str
    tag: str
    bound_type: BoundType
    bound: list

    def __post_init__(self):
        if isinstance(self.bound_type, str):
            self.bound_type = BoundType(self.bound_type)

    def to_json(self):
        """Function to convert dataclass to dict"""
        return {
            "assetId": self.asset_id,
            "tag": self.tag,
            "boundType": self.bound_type.value,
            "bound": self.bound,
        }


@dataclass
class AnnotationExportOptions:
    """Annotation exported options.

    :param split_ratio: The ratio used to split the data into training and validation sets.
    :param seed: The number used to initialize a pseudorandom
        number generator to randomize the annotation shuffling.
    :param normalized: Boolean to indicate whether the bound
        coordinates of the exported annotations should be normalized. Defaults to True.
    """

    split_ratio: float
    seed: int
    normalized: bool = True

    def to_json(self):
        """Function to convert dataclass to dict"""
        return {
            "splitRatio": self.split_ratio,
            "seed": self.seed,
            "normalized": self.normalized,
        }


@dataclass
class AnnotationExportMetadata:
    """Annotation exported metadata.

    :param format: The annotation format for bounding boxes or polygons as a string.
    :param options: The exporting options.
    """

    format: Union[AnnotationExportFormat, str]
    options: Union[AnnotationExportOptions, dict]

    def __post_init__(self):
        assert isinstance(self.format, (AnnotationExportFormat, str))
        assert isinstance(self.options, (AnnotationExportOptions, dict))

        if isinstance(self.format, str):
            self.format = AnnotationExportFormat(self.format)
        if isinstance(self.options, dict):
            self.options = AnnotationExportOptions(**self.options)

    def to_json(self):
        """Function to convert dataclass to dict"""
        return {"format": self.format.value, "options": self.options.to_json()}


@dataclass
class Pagination:
    """Pagination Params.
    If the length of the function call results exceeds the limit,
    the results will be broken into multiple pages.

    :param page: An optional cursor to specify pagination if there are multiple pages of results.
    :param limit: A limit on the number of objects to be returned in a page. Defaults to 100.
    """

    page: str = None
    limit: int = 100

    def to_json(self):
        """Function to convert dataclass to dict"""
        pagination = {
            "page": self.page,
            "limit": self.limit,
        }

        # Remove None values
        return {k: v for k, v in pagination.items() if v is not None}


@dataclass
class AssetMetadata:
    """Asset Metadata.

    :param status: The annotation status of the asset (annotated, review, completed, tofix, none).
    :param custom_metadata: A dictionary containing any key-value pairs.
    """

    status: Union[AssetStatus, None] = None
    custom_metadata: object = None

    def to_json(self):
        """Function to convert dataclass to dict"""
        asset_metadata = {
            "status": self.status,
            "customMetadata": self.custom_metadata,
        }

        # Remove None values
        return {k: v for k, v in asset_metadata.items() if v is not None}


@dataclass
class DeploymentOptions:
    """The configuration options for creating each Inference API instance.

    :param evaluation_strategy: The evaluation strategy to use
        of each Inference API, default entropy_score.
    :param evaluation_threshold: The evaluation threshold to
        use to trigger post-evaluation actions, default 0.5.
    :param evaluation_group: The asset group to which assets
        triggered by the active learning route will be uploaded, comma-separated list.
    """

    evaluation_strategy: str = None
    evaluation_threshold: float = None
    evaluation_group: List[str] = None

    def to_json(self):
        """Function to convert dataclass to dict"""
        options = {
            "evaluationStrategy": self.evaluation_strategy,
            "evaluationThreshold": self.evaluation_threshold,
            "evaluationGroup": self.evaluation_group,
        }
        # Remove None values
        return {k: v for k, v in options.items() if v is not None}


@dataclass
class DeploymentResource:
    """The resource allocation for the deployment instance, optional.
    [DEPRECATED] This field is deprecated and will be removed in future versions.

    :param cpu: The number of CPU cores to allocate, optional.
    :param ram: The amount of RAM to allocate in GB, optional.
    :param GPU_T4: The number of NVIDIA Tesla T4 GPUs to allocate, optional.
    :param GPU_L4: The number of NVIDIA Tesla L4 GPUs to allocate, optional.
    :param GPU_A100_40GB: The number of NVIDIA A100 (40GB) GPUs to allocate, optional.
    :param GPU_A100_80GB: The number of NVIDIA A100 (80GB) GPUs to allocate, optional.
    :param GPU_H100: The number of NVIDIA H100 GPUs to allocate, optional.
    """

    cpu: Optional[int] = None
    ram: Optional[int] = None
    GPU_T4: Optional[int] = None
    GPU_L4: Optional[int] = None
    GPU_A100_40GB: Optional[int] = None
    GPU_A100_80GB: Optional[int] = None
    GPU_H100: Optional[int] = None

    def __post_init__(self):
        logger.warning(
            "WARNING: The `resources` field is deprecated and will be removed in future versions. "
            "Use the `instance` field instead."
        )

    def to_json(self):
        """Function to convert dataclass to dict"""
        resource = {
            "cpu": self.cpu,
            "ram": self.ram,
            "GPU_T4": self.GPU_T4,
            "GPU_L4": self.GPU_L4,
            "GPU_A100_40GB": self.GPU_A100_40GB,
            "GPU_A100_80GB": self.GPU_A100_80GB,
            "GPU_H100": self.GPU_H100,
        }
        # Remove None values
        return {k: v for k, v in resource.items() if v is not None}


@dataclass
class DeploymentMetadata:
    """Deployment Settings Metadata.

    :param name: The name of the deployment instance.
    :param model_id: The ID of the exported artifact to be deployed.
        [DEPRECATED] This field is deprecated and will be removed in future versions.
        Use the `artifact_id` field instead.
    :param artifact_id: The ID of the artifact to be deployed.
    :param replicas: Number of deployment instances to spawn. Defaults to 1.
    :param region: The region where the deployment instance will be spawned.
    :param version_tag: The current version tag of the deployment instance.
    :param instance: The instance type for the deployment instance, optional.
    :param resources: The resource allocation for the deployment instance, optional.
        [DEPRECATED] This field is deprecated and will be removed in future versions.
    :param options: The configuration options for the deployment instance, optional.
    """

    name: Optional[str] = None
    model_id: Optional[str] = None
    artifact_id: Optional[str] = None
    replicas: int = 1
    region: Optional[str] = None
    version_tag: Optional[str] = None
    instance_id: Optional[str] = None
    resources: Optional[DeploymentResource] = None
    options: Optional[DeploymentOptions] = None

    def __post_init__(self):
        if isinstance(self.options, dict):
            self.options = DeploymentOptions(**self.options)
        if isinstance(self.resources, dict):
            logger.warning(
                "WARNING: The `resources` field is deprecated and will be removed in future versions. "
                "Use the `instance` field instead."
            )
            self.resources = DeploymentResource(**self.resources)

        if not self.instance_id and not self.resources:
            raise ValueError("Please provide either `instance_id` or `resources`.")

    def to_json(self):
        """Function to convert dataclass to dict"""
        if self.model_id is not None:
            logger.warning(
                "WARNING: The `model_id` field is deprecated and will be removed "
                "in future versions. Use the `artifact_id` field instead."
            )

        deployment_metadata = {
            "name": self.name,
            "modelId": self.model_id,
            "artifactId": self.artifact_id,
            "replicas": self.replicas,
            "region": self.region,
            "versionTag": self.version_tag,
            "instanceId": self.instance_id,
            "resources": (
                self.resources.to_json() if self.resources is not None else None
            ),
            "options": self.options.to_json() if self.options is not None else None,
        }
        return {
            k: v for k, v in deployment_metadata.items() if v is not None and v != {}
        }


@dataclass
class Accelerator:
    """The hardware accelerator to be used for the training.

    :param name: The name of the GPU to be used for the training
        (GPU_T4, GPU_P100, GPU_V100, GPU_L4, GPU_A100_40GB).
    :param count: The number of GPUs to be used for the training.
        More GPUs will use up more compute minutes. Defaults to 1.
    """

    name: str
    count: int = 1

    def to_json(self):
        """Function to convert dataclass to dict"""
        return {
            "name": self.name,
            "count": self.count,
        }


@dataclass
class Checkpoint:
    """The checkpoint metric to be used for the training.

    :param strategy: The checkpointing strategy to be used for the training.

        Checkpoint Strategies:
            STRAT_EVERY_N_EPOCH: Checkpoints are saved at intervals of n epochs.
            STRAT_ALWAYS_SAVE_LATEST: The final checkpoint is always saved.
            STRAT_LOWEST_VALIDATION_LOSS: The checkpoint with the lowest validation loss is saved.
            STRAT_HIGHEST_ACCURACY: The checkpoint with the highest accuracy is saved.

    :param metric: The checkpointing metric to be used for training.
        Note that metrics starting with "Loss" are only applicable when the
        strategy is set to "STRAT_LOWEST_VALIDATION_LOSS", and metrics starting with
        "DetectionBoxes" are only applicable when the strategy is set to "STRAT_HIGHEST_ACCURACY".

        Loss:
            Loss/total_loss
            Loss/regularization_loss
            Loss/classification_loss
            Loss/localization_loss

        Precision:
            DetectionBoxes_Precision/mAP
            DetectionBoxes_Precision/mAP@.50IOU
            DetectionBoxes_Precision/mAP@.75IOU
            DetectionBoxes_Precision/mAP (small)
            DetectionBoxes_Precision/mAP (medium)
            DetectionBoxes_Precision/mAP (large)

        Recall:
            DetectionBoxes_Recall/AR@1
            DetectionBoxes_Recall/AR@10
            DetectionBoxes_Recall/AR@100
            DetectionBoxes_Recall/AR@100 (small)
            DetectionBoxes_Recall/AR@100 (medium)
            DetectionBoxes_Recall/AR@100 (large)

    :param evaluation_interval: The step interval for checkpoint
        evaluation during training. Defaults to 250.
    """

    strategy: str
    metric: str = None
    evaluation_interval: int = 250

    def to_json(self):
        """Function to convert dataclass to dict"""

        checkpoint = {
            "strategy": self.strategy,
            "metric": self.metric,
            "evaluationInterval": self.evaluation_interval,
        }
        # Remove None values
        return {k: v for k, v in checkpoint.items() if v is not None}


@dataclass
class Limit:
    """The limit configuration for the training.

    :param metric: The limit metric for the training.

        Limit Metrics:
            LIM_MINUTE: Limits the training to a maximum number of minutes before it is killed.
            LIM_EPOCHS: Limits the training to a maximum number of epochs before it is killed.
            LIM_NONE: No limit will be set for the training.

    :param value: The limit value for the training. This value will not be
        used if the limit metric is "LIM_NONE". Defaults to 1.
    """

    metric: str
    value: int = 1

    def to_json(self):
        """Function to convert dataclass to dict"""
        return {
            "metric": self.metric,
            "value": self.value,
        }


@dataclass
class RunSetupMetadata:
    """The settings to start training.

    :param accelerator: The hardware accelerator to be used for the training.
    :param checkpoint: The checkpoint metric to be used for the training.
    :param limit: The limit configuration for the training.
    :param preview: [DEPRECATED] Boolean to indicate whether preview
        is enabled for the training. Use the `advancedEvaluation` field instead.
        Defaults to True.
    :param matrix: [DEPRECATED] Boolean to indicate whether confusion matrix
        is enabled for the training. Use the `advancedEvaluation` field instead.
        Defaults to True.
    :param advanced_evaluation: Boolean to indicate whether advanced evaluation
        features are enabled for the training.
    """

    accelerator: Accelerator = None
    checkpoint: Checkpoint = None
    limit: Limit = None
    preview: Optional[bool] = None
    matrix: Optional[bool] = None
    advanced_evaluation: Optional[bool] = None
    debug: bool = False

    def __post_init__(self):
        if isinstance(self.accelerator, dict):
            self.accelerator = Accelerator(**self.accelerator)
        if isinstance(self.checkpoint, dict):
            self.checkpoint = Checkpoint(**self.checkpoint)
        if isinstance(self.limit, dict):
            self.limit = Limit(**self.limit)

        if self.preview is None:
            self.preview = (
                self.advanced_evaluation
                if self.advanced_evaluation is not None
                else True
            )
        else:
            logger.warning(
                "WARNING: Preview and other advanced evaluation features will be merged into a "
                "single field `advanced_evaluation` in future versions. The `preview` field is "
                "deprecated and will be removed in future versions."
            )

        if self.matrix is not None:
            logger.warning(
                "WARNING: Confusion matrix will be enabled by default and is no longer billable. "
                "The `matrix` field is deprecated and will be removed in future versions."
            )


@dataclass
class FlowMetadata:
    """Workflow Metadata.

    :param title: The title of the workflow.
    """

    title: str

    def to_json(self):
        """Function to convert dataclass to dict"""
        return {
            "title": self.title,
        }


@dataclass
class AssetFilter:
    """Asset list filter."""

    groups: List[str] = None
    filename: str = None
    metadata_query: str = None
    status: Union[AssetStatus, str, None] = None

    def __post_init__(self):
        assert isinstance(self.groups, (List, type(None)))

        if isinstance(self.status, str):
            self.status = AssetStatus(self.status)

    def to_json(self):
        """Function to convert dataclass to dict"""
        asset_filters = {
            "group": self.groups,
            "filename": self.filename,
            "metadataQuery": self.metadata_query,
            "status": self.status.value if self.status is not None else None,
        }

        return {k: v for k, v in asset_filters.items() if v is not None}


@dataclass
class AnnotationFilter:
    """Annotation list filter."""

    asset_ids: List[str] = None

    def __post_init__(self):
        assert isinstance(self.asset_ids, (List, type(None)))

    def to_json(self):
        """Function to convert dataclass to dict"""
        annotation_filters = {
            "assetId": self.asset_ids,
        }

        return {k: v for k, v in annotation_filters.items() if v is not None}


@dataclass
class TagMetadata:
    """Tag Metadata."""

    name: str = None
    color: str = None
    description: str = None

    def to_json(self):
        """Function to convert dataclass to dict"""
        tag = {
            "name": self.name,
            "color": self.color,
            "description": self.description,
        }

        # Remove None values
        return {k: v for k, v in tag.items() if v is not None}


@dataclass
class ArtifactExportOptions:
    """Artifact Export Options."""

    format: Union[ModelFormat, str]
    quantization: str = None

    def __post_init__(self):
        assert isinstance(self.format, (ModelFormat, str))

        if isinstance(self.format, str):
            self.format = ModelFormat(self.format)

    def to_json(self):
        """Function to convert dataclass to dict"""
        options = {
            "format": self.format.value,
            "quantization": self.quantization,
        }

        # Remove None values
        return {k: v for k, v in options.items() if v is not None}


@dataclass
class ArtifactFilters:
    """Artifact list filter."""

    run_ids: List[str] = None

    def __post_init__(self):
        assert isinstance(self.run_ids, (List, type(None)))

    def to_json(self):
        """Function to convert dataclass to dict"""
        artifact_filters = {
            "runId": self.run_ids,
        }

        return {k: v for k, v in artifact_filters.items() if v is not None}



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   mock.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Test Mock Cases
"""


class MockResponse:
    """Datature Mock Request Response."""

    def __init__(self, json_data, status_code):
        self.json_data = json_data
        self.status_code = status_code

    def json(self):
        """Request json function."""

        return self.json_data



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   version.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   SDK version module
"""

__version__ = "1.9.1b1"
